/***********************************************************************/
/* FreeVR OpenGL tutorial example 0: ex0_xwin.c                        */
/* Last tested with: N/A (since this is not a VR program)              */
/* Last Modified: December 3, 2009                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/*                                                                     */
/* This example code demonstrates an OpenGL application program that   */
/*   displays a blue cube and a red pyramid in a simple X window.      */
/*                                                                     */
/*                                                                     */
/* The features of this example include:                               */
/*   - Opens an X window                                               */
/*   - Creates an OpenGL viewport                                      */
/*   - Renders a simple static world                                   */
/*                                                                     */
/*                                                                     */
/* Considerations:                                                     */
/*   - A lot of code is necessary to open the GLX window (X w/ OpenGL) */
/*   - The main loop would typically watch for many more X Events      */
/*   - This is not a VR program                                        */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */
#include <GL/glx.h>	/* needed for X-windows interfacing to OpenGL */

void	exit(int status);
void	draw_world(void);			/* graphically render the world    */

#define WINDOW_WIDTH	300
#define WINDOW_HEIGHT	300

static int      snglBuf[] = {GLX_RGBA, GLX_DEPTH_SIZE, 16, None};
static int      dublBuf[] = {GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, None};
static int      quadBuf[] = {GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, GLX_STEREO, None};

Display        *dpy;
Window          win;
GLboolean       doubleBuffer = GL_TRUE;
GLboolean       stereoBuffer = GL_TRUE;


/***********************************************************************/
main(int argc, char* argv[])
{
	XVisualInfo    *vi;
	Colormap        cmap;
	XSetWindowAttributes swa;
	GLXContext      cx;
	XEvent          event;
	int             dummy;

	/**********************************/
	/***  initialize GLX routines   ***/
	/**********************************/
	/*** open a connection to the X server ***/
	dpy = XOpenDisplay(NULL);
	if (dpy == NULL) {
		perror("could not open display");
		exit(0);
	}

	/*** make sure OpenGL's GLX extension supported ***/
	if (!glXQueryExtension(dpy, &dummy, &dummy)) {
		perror("X server has no OpenGL GLX extension");
		exit(0);
	}

	/*** find an appropriate visual ***/
	/* find an OpenGL-capable RGB visual with depth buffer */
	vi = glXChooseVisual(dpy, DefaultScreen(dpy), quadBuf);
	if (vi == NULL) {
		printf("Unable to acquire an X11 visual with stereo (quad) buffering, trying for mono-double.\n");
		vi = glXChooseVisual(dpy, DefaultScreen(dpy), dublBuf);
		if (vi == NULL) {
			printf("Unable to acquire an X11 visual with double buffering, trying for single.\n");
			vi = glXChooseVisual(dpy, DefaultScreen(dpy), snglBuf);
			if (vi == NULL) {
				perror("no RGB visual with depth buffer");
				exit(0);
			}
			doubleBuffer = GL_FALSE;
		}
		stereoBuffer = GL_FALSE;
	}
	if (vi->class != TrueColor) {
		perror("TrueColor visual required for this program");
		exit(0);
	}

	/*** create an OpenGL rendering context  ***/
	/* create an OpenGL rendering context */
	cx = glXCreateContext(dpy, vi, /* no sharing of display lists */ None,
			       /* direct rendering if possible */ GL_TRUE);
	if (cx == NULL) {
		perror("could not create rendering context");
		exit(0);
	}

	/*** create an X window with the selected visual ***/
	/* create an X colormap since probably not using default visual */
	cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual, AllocNone);
	swa.colormap = cmap;
	swa.border_pixel = 0;
	swa.event_mask = ExposureMask | ButtonPressMask | StructureNotifyMask;
	win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, 0, vi->depth,
			    InputOutput, vi->visual, CWBorderPixel | CWColormap | CWEventMask, &swa);
	XSetStandardProperties(dpy, win, "simple-glx", "simple-glx", None, argv, argc, NULL);

	/*** bind the rendering context to the window ***/
	glXMakeCurrent(dpy, win, cx);

	/*** request the X window to be displayed on the screen ***/
	XMapWindow(dpy, win);

	/*** configure the OpenGL context for rendering ***/
	glEnable(GL_DEPTH_TEST);/* enable depth buffering */
	glDepthFunc(GL_LESS);	/* pedantic, GL_LESS is the default */
	glClearDepth(1.0);	/* pedantic, 1.0 is the default */
	/* frame buffer clears should be to black */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	/* set up projection transform */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1.0, 1.0, -1.0, 1.0, 1.0, 10.0);
	/* establish initial viewport */
	glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);	/* pedantic, full window size is default viewport */

	/***************************/
	/*** do world simulation ***/
	/***************************/
	while (1) {
		/*** Get and handle inputs ***/
		do {
			/* Note, the do-while form will always enter the loop even */
			/*   when no events are waiting, and will then pause at    */
			/*   the XNextEvent() function, which blocks.              */
			XNextEvent(dpy, &event);
			switch (event.type) {
			case ConfigureNotify:
				glViewport(0, 0, event.xconfigure.width, event.xconfigure.height);
			}
		} while (XPending(dpy));

		/*** Handle view placement ***/
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0.0, -3.0, 0.0);

		/*** Render the world ***/
		draw_world();

		/*** Swap buffers (or flush GL queue) ***/
		if (doubleBuffer)
			glXSwapBuffers(dpy, win);
		else	glFlush();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from shapes.c */
void	draw_cube();
void	draw_pyramid();


/*************************************************************/
/* draw_world(): Render the two objects in the world -- the  */
/*   blue cube, and red pyramid -- such that they are on the */
/*   XZ plane (typically the floor), and 6 feet forward      */
/*   (typically on the other side of a CAVE's front screen). */
/*************************************************************/
void draw_world(void)
{
	/**************************************/
	/* set the polygon shading parameters */
	glShadeModel(GL_SMOOTH);
	glLineWidth(2.5);

	/**************************************/
	/* clear the screen -- very important */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/************************/
	/* draw all the objects */

	/* a blue cube */
	glColor3ub(100, 100, 255);
	glTranslatef(-2.0, 1.0, -6.0);
	draw_cube();
	glColor3ub(0, 0, 0);
	draw_cube_outline();

	/* a red pyramid */
	glColor3ub(255, 100, 100);
	glTranslatef(4.0, 0.0, 0.0);
	draw_pyramid();
	glColor3ub(0, 0, 0);
	draw_pyramid_outline();
}

