/***********************************************************************/
/* FreeVR OpenGL tutorial example 14e: ex14e_joyst.c                   */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: January 11, 2007                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that        */
/*   allows the user to use special objects in the virtual world.      */
/*   Specifically, this example introduces a joystick controler widget */
/*   that looks similar to the lever.  However now two valuator values */
/*   can be specified.  The joystick is moved by pressing the left     */
/*   while contacting the ball handle to grab it.                      */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - nothing new provided by the library.                            */
/*                                                                     */
/* Considerations:                                                     */
/*                                                                     */
/* This application uses the files:                                    */
/*   - wrs_shapes.c  (Bill Sherman's simple OpenGL shapes)             */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() & getenv() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for fabs() & trig functions */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */

/* Define some values to adjust the compilation of this example */
#define	DRAW_UITEXT
#define	USE_SOCKETS
#define MAX_OBJECTS	1000
#define RADIO_SEP	  0.7
#define	SLIDER_LENGTH	  3.0
#define	LEVER_MAX_ANGLE	 50.0
#define	LEVER_KNOB_SIZE   0.5
#define	LEVER_ARM_LENGTH  1.0
#undef	LEVER_SNAPBACK
#define	JOYST_MAX_ANGLE	 40.0
#define	JOYST_KNOB_SIZE   0.5
#define	JOYST_ARM_LENGTH  0.7
#undef	JOY_SNAPBACK


/* Define a structure for information about each individual object. */
typedef struct {
		float			x, y, z;		/* location of the object     */
		float			size;			/* size of the object         */
		GLubyte			color[3];		/* color of the object        */
		enum {	cube,
			sphere,
			pyramid,
			ud_pyramid }	shape;			/* shape of the object        */
		enum {	vw_normal,
			ui_button,
			ui_radio,
			ui_slider,
			ui_lever,
			ui_joyst }	type;			/* type of the object         */
		int			ui_button_value;	/* button value of a widget   */
		float			ui_valuator_value;	/* valuator value of a widget */
		float			ui_valuator2_value;	/* valuator value of a widget */
		int			ui_num_buttons;		/* number of radio buttons    */
		char			*ui_radio_names[10];	/* must be allocated in shared memory */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		quit_bit;			/* a flag to pass along the desire to terminate */
		int		number_of_objects;		/* number of objects in world */
		int		contacted_object;		/* object contacted by wand   */
		int		last_contacted_object;		/* object previously contacted*/
		ObjectType	object[MAX_OBJECTS];		/* array of objects in world  */
		vrLock		lock;				/* to avoid bad multiprocess accesses */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations     */

void	init_gfx();				/* initialize the graphics           */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world   */
void	init_world(WorldDataType *wd);		/* initialize the world parameters   */
void	update_world(WorldDataType *wd);	/* perform the world simulation      */

void	end_world();				/* close down the world operations   */
void	init_objects(WorldDataType *wd);	/* initalize the objects in the world*/


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	wd->lock = vrLockCreate();
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("example 14e -- joystick widget");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to demonstrate a joystick virtual control");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Move the selected object");
	vrInputSet2switchDescription(3, "Reset the travel matrix (take us home)");
	vrInputSetValuatorDescription(0, "Rotate us in the world about the Y-axis");
	vrInputSetValuatorDescription(1, "Move us though the world in the direction we point");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a short cyan pointer to manipulate objects & choose travel dir");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	vrUserTravelReset(VR_ALLUSERS);
	/* run until the user has indicated it's time not to */
	while(!wd->quit_bit) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	end_world();

	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

#ifdef USE_SOCKETS /* { */
static int	listen_sock = -1;	/* socket to receive commands */
					/*   listen_sock is not in wd        */
					/*   because it needs to be          */
					/*   available to exit_application() */
					/*   which can have no arguments.    */
#endif /* } USE_SOCKETS */


/****************************************************************/
/* init_world(): create the initial conditions of the dynamic   */
/*   simulation.  If the "VR_PORT" environment variable is set, */
/*   then its value is used as the initial server socket port.  */
/*   Otherwise, the arbitrarily chosen port number 4000 is used.*/
/*   Since IRIX, and perhaps other Unicies have a "feature" in  */
/*   which a socket port isn't usable again until after 60      */
/*   seconds have passed since the connection was closed, if the*/
/*   initial socket isn't available, we try the next two ports  */
/*   in succession.  The selected port is reported to the tty.  */
/****************************************************************/
void init_world(WorldDataType *wd)
{
#ifdef USE_SOCKETS /* { */
	int	inet_port;
	int	inet_port_max;
	char	host_machine[128];
#endif /* } USE_SOCKETS */

	/*************************/
	/* set world state flags */
	vrLockWriteSet(wd->lock);
	wd->quit_bit = 0;
	vrLockWriteRelease(wd->lock);

	/***************************/
	/* setup the world content */
	init_objects(wd);

#ifdef USE_SOCKETS /* { */
	/******************************/
	/* initialize the socket port */
	gethostname(host_machine, 128);

	/* loop through N possible internet ports to hopefully find one that works */
	if (getenv("VR_PORT") != NULL)
		inet_port = atoi(getenv("VR_PORT"));
	else	inet_port = 4000;
	for (inet_port_max = inet_port + 2; listen_sock < 0 && inet_port <= inet_port_max; inet_port++) {
		printf("Port %d -- ", inet_port);
		fflush(stdout);
		listen_sock = vrSocketCreateListen(&inet_port, 0);
	}
	inet_port--;
	if (listen_sock < 0)
		printf("WARNING: Unable to find an open socket port.\n");
	else	printf("\rConnect to '%s:%d' to send commands to the simulation\n", host_machine, inet_port);
#endif /* } USE_SOCKETS */
}


/***************************************************************/
/* init_objects(): place objects into the world data structure */
/*   at some initial positions.                                */
/***************************************************************/
void init_objects(WorldDataType *wd)
{
	int	objnum = 0;

	vrLockWriteSet(wd->lock);

	/* object 1 - a blue cube */
	wd->object[objnum].type = vw_normal;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x =  0.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 2 - a red pyramid */
	wd->object[objnum].type = vw_normal;
	wd->object[objnum].shape = pyramid;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 3 - an upside-down green pyramid */
	wd->object[objnum].type = vw_normal;
	wd->object[objnum].shape = ud_pyramid;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 4 - a user interface button */
	wd->object[objnum].type = ui_button;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 0.4;
	wd->object[objnum].ui_button_value = 1;
	wd->object[objnum].color[0] = 200;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  4.0;
	wd->object[objnum].z = -2.0;
	objnum++;

	/* object 5 - a user interface slider */
	wd->object[objnum].type = ui_slider;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 0.4;
	wd->object[objnum].ui_valuator_value = 1.0;
	wd->object[objnum].color[0] = 208;
	wd->object[objnum].color[1] = 113;
	wd->object[objnum].color[2] =  33;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  4.0;
	wd->object[objnum].z = -1.0;
	objnum++;

	/* object 6 - a user interface lever */
	wd->object[objnum].type = ui_lever;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].ui_valuator_value = 0.0;
	wd->object[objnum].color[0] = 101;
	wd->object[objnum].color[1] = 167;
	wd->object[objnum].color[2] = 107;
	wd->object[objnum].x = -1.5;
	wd->object[objnum].y =  3.5;
	wd->object[objnum].z = -3.0;
	objnum++;

	/* object 7 - a user interface radio button */
	wd->object[objnum].type = ui_radio;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 0.4;
	wd->object[objnum].ui_button_value = 0;
	wd->object[objnum].color[0] = 212;	/* rose */
	wd->object[objnum].color[1] =  42;
	wd->object[objnum].color[2] =  95;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z =  1.0;
	wd->object[objnum].ui_num_buttons = 4;
	wd->object[objnum].ui_radio_names[0] = vrShmemStrDup("choice 1");
	wd->object[objnum].ui_radio_names[1] = vrShmemStrDup("choice 2");
	wd->object[objnum].ui_radio_names[2] = vrShmemStrDup("choice 3");
	wd->object[objnum].ui_radio_names[3] = vrShmemStrDup("choice 4");
	objnum++;

	/* object 8 - a user interface joystick */
	wd->object[objnum].type = ui_joyst;
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 0.5;
	wd->object[objnum].ui_valuator_value  = 0.0;
	wd->object[objnum].ui_valuator2_value = 0.0;
	wd->object[objnum].color[0] =  74;
	wd->object[objnum].color[1] = 103;
	wd->object[objnum].color[2] = 130;
	wd->object[objnum].x =  1.5;
	wd->object[objnum].y =  3.5;
	wd->object[objnum].z = -3.0;
	objnum++;

	wd->number_of_objects = objnum;
	wd->contacted_object = -1;
	wd->last_contacted_object = wd->contacted_object;

	/* sanity test */
	if (objnum > MAX_OBJECTS) {
		printf("DANGER: Number of objects (%d), exceeds maximum allowable (%d).  You may experience issues.\n", objnum, MAX_OBJECTS);
	}

	vrLockWriteRelease(wd->lock);
}


/*******************************************************/
/* end_world(): clean up anything opened by simulation */
/*******************************************************/
void end_world()
{
#ifdef USE_SOCKETS /* { */
	vrSocketClose(listen_sock);
#endif /* } USE_SOCKETS */
}


/*********************************************************************/
/* update_world(): In this version, we handle the user interactions. */
/*   Both selecting, and moving the contacted object.                */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a scale factor to tune travel movement               */
#define TURN_FACTOR	25.0				/* a scale factor to tune travel movement               */
#define SUB_JS_EPSILON(a)	((a)-copysign(JS_EPSILON,(a)))	/* a macro to remove the joystick epsilon range */
#ifdef USE_SOCKETS /* { */
static	char	*help_msg = "Socket commands:\n"
				"\thelp -- this message\n"
				"\tquit -- quit the application\n"
				"\tclose -- close the connection\n"
				"\treset -- reset the application\n"
				"\tbutton? -- report the value of the UI button\n"
				"\tbutton on -- set the value of the UI button to 'on'\n"
				"\tbutton off -- set the value of the UI button to 'off'\n"
				"\tradio? -- report the value of the UI radio-button\n"
				"\tradio <n> -- set the value of the UI radio-button\n"
				"\tslider? -- report the value of the UI slider\n"
				"\tslider <n> -- set the value of the UI slider\n"
				"\tlever? -- report the value of the UI lever\n"
				"\tlever <n> -- set the value of the UI lever\n"
				"\tjoyst? -- report the values of the UI joystick\n"
				"\tjoyst1 <n> -- set X value of the UI joystick\n"
				"\tjoyst2 <n> -- set Y value of the UI joystick\n"
				"\tmove cube up -- translate the cube up 2 feet\n"
				"\tmove cube down -- translate the cube down 2 feet\n"
				"\twhere -- report the user's position in the world\n"
				"\twhere am i -- report the user's position in the world\n"
				"\tshow ActiveWalls -- report some CAVE config info\n"
		"";
static	int	cmd_socket = -1;	/* socket to get commands from */
static	char	buf[128];		/* socket read buffer          */
static	char	msg[128];		/* socket send buffer          */
	int	result;			/* result of reading a socket  */
#endif /* } USE_SOCKETS */

static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
static	vrPoint		last_wand_vw_locpnt;		/* VW location of the wand in the last update frame */
	vrPoint		wand_vw_locpnt;			/* current location of the wand (in VW coords) */
	vrPoint		wand_rw_locpnt;			/* current location of the wand (in RW coords) */
#define	last_wand_vw_loc last_wand_vw_locpnt.v
#define	wand_vw_location wand_vw_locpnt.v
#define	wand_rw_location wand_rw_locpnt.v
	vrVector	wand_rw_pointvec;		/* current vector pointing out of the wand (in RW coords) */
	double		delta_time;			/* time since last update (in seconds) */
	double		delta_move;			/* distance an object should be moved */
	double		joy_x, joy_y;			/* Joystick values */
	int		touch_obj = -1;			/* the currently touched object   */
	int		objnum;				/* for looping throuh the objects */
	ObjectType	*object;			/* pointer to object to be compared/moved  */
	float		size;				/* size of object -- used for isect comparison */
	int		button_value;
	float		slider_value;
	float		joyst1_value;
	float		joyst2_value;
	float		value_sin;
	float		value_cos;
	float		value2_sin;
	float		value2_cos;

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/**********************************/
	/** Handle Manipulation via wand **/
	/**********************************/

	/**********************************************************/
	/* check for contact between the wand and all the objects */
	vrLockReadSet(wd->lock);

	vrPointGetVWFromUser6sensor(&wand_vw_locpnt, 0, WAND_SENSOR);
	vrPointGetRWFrom6sensor(&wand_rw_locpnt, WAND_SENSOR);

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);
		size = object->size;

		switch (object->type) {
		case vw_normal:	/* determine contact with normal object */
			if ( (wand_vw_location[VR_X] < object->x + size)
			  && (wand_vw_location[VR_X] > object->x - size)
			  && (wand_vw_location[VR_Y] < object->y + size)
			  && (wand_vw_location[VR_Y] > object->y - size)
			  && (wand_vw_location[VR_Z] < object->z + size)
			  && (wand_vw_location[VR_Z] > object->z - size) ) {
				touch_obj = objnum;
			}
			break;
		
		case ui_button:	/* determine contact with button widget */
			if ( (wand_rw_location[VR_X] < object->x + size)
			  && (wand_rw_location[VR_X] > object->x - size)
			  && (wand_rw_location[VR_Y] < object->y + size)
			  && (wand_rw_location[VR_Y] > object->y - size)
			  && (wand_rw_location[VR_Z] < object->z + size)
			  && (wand_rw_location[VR_Z] > object->z - size) ) {
				touch_obj = objnum;
			}
			break;

		case ui_radio: /* determine contact with radio-button widget */
			if ( (wand_rw_location[VR_X] < object->x + size)
			  && (wand_rw_location[VR_X] > object->x - size)
			  && (wand_rw_location[VR_Y] < object->y + (object->ui_num_buttons/2.0 * (size * (2.0 + RADIO_SEP))))
			  && (wand_rw_location[VR_Y] > object->y - (object->ui_num_buttons/2.0 * (size * (2.0 + RADIO_SEP))))
			  && (wand_rw_location[VR_Z] < object->z + size)
			  && (wand_rw_location[VR_Z] > object->z - size) ) {
				touch_obj = objnum;
			}
			break;

		case ui_slider: /* determine contact with slider widget */
			/* check intersection with the (moving) grab bar */
			/* NOTE: this crudely assumes vertically aligned slider */
			if (
				/* contact with the grab-plate */
			    ((wand_rw_location[VR_X] < object->x + size)
			  && (wand_rw_location[VR_X] > object->x - size)
			  && (wand_rw_location[VR_Y] < (object->y + object->ui_valuator_value * size * SLIDER_LENGTH) + size * 0.2)
			  && (wand_rw_location[VR_Y] > (object->y + object->ui_valuator_value * size * SLIDER_LENGTH) - size * 0.2)
			  && (wand_rw_location[VR_Z] < object->z + size)
			  && (wand_rw_location[VR_Z] > object->z - size))
			 ||	/* or contact with the slider-bar */
			    ((wand_rw_location[VR_X] < object->x + size * 0.2)
			  && (wand_rw_location[VR_X] > object->x - size * 0.2)
			  && (wand_rw_location[VR_Y] < object->y + size * SLIDER_LENGTH)
			  && (wand_rw_location[VR_Y] > object->y - size * SLIDER_LENGTH)
			  && (wand_rw_location[VR_Z] < object->z + size * 0.2)
			  && (wand_rw_location[VR_Z] > object->z - size * 0.2))
				) {
				touch_obj = objnum;
			}
			break;

		case ui_lever: /* determine contact with lever widget */
			/* check intersection with the (moving) lever knob */
			/* NOTE: this crudely assumes horizontally aligned lever along the X-axis */

			/* calculate the sine & cosine for the tip of the lever */
			value_sin = sinf(object->ui_valuator_value * LEVER_MAX_ANGLE * (3.1415 / 180.0));
			value_cos = cosf(object->ui_valuator_value * LEVER_MAX_ANGLE * (3.1415 / 180.0));
			if ( (wand_rw_location[VR_X] < (object->x + (value_sin * LEVER_ARM_LENGTH * size * 2.0) + size * LEVER_KNOB_SIZE))
			  && (wand_rw_location[VR_X] > (object->x + (value_sin * LEVER_ARM_LENGTH * size * 2.0) - size * LEVER_KNOB_SIZE))
			  && (wand_rw_location[VR_Y] < (object->y                                               + size * LEVER_KNOB_SIZE))
			  && (wand_rw_location[VR_Y] > (object->y                                               - size * LEVER_KNOB_SIZE))
			  && (wand_rw_location[VR_Z] < (object->z + (value_cos * LEVER_ARM_LENGTH * size * 2.0) + size * LEVER_KNOB_SIZE))
			  && (wand_rw_location[VR_Z] > (object->z + (value_cos * LEVER_ARM_LENGTH * size * 2.0) - size * LEVER_KNOB_SIZE)) ) {
				touch_obj = objnum;
			} else {
#ifdef LEVER_SNAPBACK
				object->ui_valuator_value  =  0.0;
#endif
			}
			break;

		case ui_joyst: /* determine contact with joystick widget */
			/* check intersection with the (moving) joystick knob */
			/* NOTE: this crudely assumes aligned joystick along the XZ-plane */

			/* calculate the sines & cosines for the tip of the stick */
			value_sin  = sinf(object->ui_valuator_value  *  JOYST_MAX_ANGLE * (3.1415 / 180.0));
			value_cos  = cosf(object->ui_valuator_value  *  JOYST_MAX_ANGLE * (3.1415 / 180.0));
			value2_sin = sinf(object->ui_valuator2_value * -JOYST_MAX_ANGLE * (3.1415 / 180.0));
			value2_cos = cosf(object->ui_valuator2_value * -JOYST_MAX_ANGLE * (3.1415 / 180.0));
			if ( (wand_rw_location[VR_X] < (object->x + ((value_sin * value2_cos) * JOYST_ARM_LENGTH * size * 2.0) + size * JOYST_KNOB_SIZE))
			  && (wand_rw_location[VR_X] > (object->x + ((value_sin * value2_cos) * JOYST_ARM_LENGTH * size * 2.0) - size * JOYST_KNOB_SIZE))
			  && (wand_rw_location[VR_Y] < (object->y + ((value_cos * value2_cos) * JOYST_ARM_LENGTH * size * 2.0) + size * JOYST_KNOB_SIZE))
			  && (wand_rw_location[VR_Y] > (object->y + ((value_cos * value2_cos) * JOYST_ARM_LENGTH * size * 2.0) - size * JOYST_KNOB_SIZE))
			  && (wand_rw_location[VR_Z] < (object->z + ((value2_sin * 1.0)       * JOYST_ARM_LENGTH * size * 2.0) + size * JOYST_KNOB_SIZE))
			  && (wand_rw_location[VR_Z] > (object->z + ((value2_sin * 1.0)       * JOYST_ARM_LENGTH * size * 2.0) - size * JOYST_KNOB_SIZE)) ) {
				touch_obj = objnum;
			} else {
#ifdef JOY_SNAPBACK
				object->ui_valuator_value  =  0.0;
				object->ui_valuator2_value =  0.0;
#endif
			}
			break;
		}
	}
	vrLockReadRelease(wd->lock);


	/*************************************************************/
	/* perform type-appropriate actions for the contacted object */

	/* NOTE: ideally we would use write-locks specific to each */
	/*   object, and only where that object is being changed.  */
	vrLockWriteSet(wd->lock);	/* lock while changing an object's location */
	wd->last_contacted_object = wd->contacted_object;
	wd->contacted_object = touch_obj;
	if (wd->contacted_object != -1) {
		object = &(wd->object[wd->contacted_object]);
		size = object->size;

		switch (object->type) {
		case vw_normal: /* perform manipulation on normal object */
			/* when left wand button is pressed, a contacted VW object can be moved */
			if (vrGet2switchValue(1)) {

				object->x += (wand_vw_location[VR_X] - last_wand_vw_loc[VR_X]);
				object->y += (wand_vw_location[VR_Y] - last_wand_vw_loc[VR_Y]);
				object->z += (wand_vw_location[VR_Z] - last_wand_vw_loc[VR_Z]);
			}
			break;

		case ui_button: /* perform manipulation on button widget */
			/* this code only needed for direct-control of button */

			/* CONTACT-trigger: when we begin to touch a button, toggle its state */
			if (wd->contacted_object != wd->last_contacted_object) {
				object->ui_button_value ^= 1;
			}
			break;

		case ui_radio: { /* perform manipulation on radio-button widget */
			/* this code only needed for direct-control of button */
			int	count;

			/* CONTACT-trigger: when we begin to touch a button, toggle its state */
			if (1) {
				for (count = 0; count < object->ui_num_buttons; count++) {

					if ( (wand_rw_location[VR_X] < object->x + size)
					  && (wand_rw_location[VR_X] > object->x - size)
					  && (wand_rw_location[VR_Y] < object->y + ((object->ui_num_buttons - 1)/2.0 * (size * (2.0 + RADIO_SEP))) - count * (size * (2.0 + RADIO_SEP)) + size)
					  && (wand_rw_location[VR_Y] > object->y + ((object->ui_num_buttons - 1)/2.0 * (size * (2.0 + RADIO_SEP))) - count * (size * (2.0 + RADIO_SEP)) - size)
					  && (wand_rw_location[VR_Z] < object->z + size)
					  && (wand_rw_location[VR_Z] > object->z - size) ) {
						object->ui_button_value = count;
					}
				}
			}
		 }
			break;

		case ui_slider: /* perform manipulation on slider widget */
			/* this code only needed for direct-control of slider */
			/* NOTE: this crudely assumes vertically aligned slider */

			/* BUTTON-trigger: when left button is pressed, grab bar can be moved */
			/* NOTE: should also have a contact-trigger version */
			if (vrGet2switchValue(1)) {
				object->ui_valuator_value = (wand_rw_location[VR_Y] - object->y) / (size * SLIDER_LENGTH);
				/* put limits on how far the grab-bar moves */
				if (object->ui_valuator_value > 1.0)
					object->ui_valuator_value = 1.0;
				if (object->ui_valuator_value < -1.0)
					object->ui_valuator_value = -1.0;
			}
			break;

		case ui_lever: /* perform manipulation on lever widget */
			/* this code only needed for direct-control of lever */
			/* NOTE: this crudely assumes horizontally aligned lever along the X-axis */

			/* BUTTON-trigger: when left button is pressed, lever-knob can be moved */
			/* NOTE: should also have a contact-trigger version */
			if (vrGet2switchValue(1)) {
				object->ui_valuator_value = atanf((wand_rw_location[VR_X] - object->x) / (wand_rw_location[VR_Z] - object->z)) * (180.0 / 3.1415) / LEVER_MAX_ANGLE;

				/* put limits on how far the grab-bar moves */
				if (object->ui_valuator_value > 1.0)
					object->ui_valuator_value = 1.0;
				if (object->ui_valuator_value < -1.0)
					object->ui_valuator_value = -1.0;
			} else {
#ifdef LEVER_SNAPBACK
				object->ui_valuator_value  =  0.0;
#endif
			}
			break;

		case ui_joyst: /* perform manipulation on joystick widget */
			/* this code only needed for direct-control of joystick */
			/* NOTE: this crudely assumes aligned joystick along the XZ-plane */

			/* BUTTON-trigger: when left button is pressed, lever-knob can be moved */
			/* NOTE: should also have a contact-trigger version */
			if (vrGet2switchValue(1)) {
				object->ui_valuator_value  =  atanf((wand_rw_location[VR_X] - object->x) / (wand_rw_location[VR_Y] - object->y)) * (180.0 / 3.1415) / LEVER_MAX_ANGLE;
				object->ui_valuator2_value = -atanf((wand_rw_location[VR_Z] - object->z) / (wand_rw_location[VR_Y] - object->y)) * (180.0 / 3.1415) / LEVER_MAX_ANGLE;

				/* put limits on how far the grab-bar moves */
				if (object->ui_valuator_value > 1.0)
					object->ui_valuator_value = 1.0;
				if (object->ui_valuator_value < -1.0)
					object->ui_valuator_value = -1.0;
				if (object->ui_valuator2_value > 1.0)
					object->ui_valuator2_value = 1.0;
				if (object->ui_valuator2_value < -1.0)
					object->ui_valuator2_value = -1.0;
			} else {
#ifdef JOY_SNAPBACK
				object->ui_valuator_value  =  0.0;
				object->ui_valuator2_value =  0.0;
#endif
			}
			break;
		}
	}
	vrLockWriteRelease(wd->lock);

	/******************************************************************/
	/* store the current wand location for next time through the loop */
	last_wand_vw_loc[VR_X] = wand_vw_location[VR_X];
	last_wand_vw_loc[VR_Y] = wand_vw_location[VR_Y];
	last_wand_vw_loc[VR_Z] = wand_vw_location[VR_Z];


	/************************************************/
	/** Handle Travel via wand and virtual devices **/
	/************************************************/

#if 0 /* set to one to demonstrate virtual device control of the joystick (itself a virtual device) */
	vrLockWriteSet(wd->lock);
	/* set the joystick's position from the slider and lever -- virtual device control */
	wd->object[7].ui_valuator_value = wd->object[4].ui_valuator_value;
	wd->object[7].ui_valuator2_value = wd->object[5].ui_valuator_value;
	vrLockWriteRelease(wd->lock);
#endif

	vrLockReadSet(wd->lock);	/* lock while getting the UI values */
	button_value = wd->object[3].ui_button_value;
	slider_value = wd->object[4].ui_valuator_value;
	joyst1_value = wd->object[7].ui_valuator_value;
	joyst2_value = wd->object[7].ui_valuator2_value;
	vrLockReadRelease(wd->lock);

	/* only travel if UI-button is on */
	if (button_value) {
		/* pressing the right wand button resets us to the initial position */
		if (vrGet2switchValue(3)) {
			vrUserTravelReset(VR_ALLUSERS);
		}

		/* use wand joystick to fly through world */
		joy_x = joyst1_value;
		joy_y = joyst2_value;

		if (fabs(joy_x) > JS_EPSILON)
			vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (delta_time * SUB_JS_EPSILON(joy_x) * -TURN_FACTOR * slider_value));

		if (fabs(joy_y) > JS_EPSILON) {
			delta_move = delta_time * SUB_JS_EPSILON(joy_y) * MOVE_FACTOR * slider_value;
			vrVectorGetRWFrom6sensorDir(&wand_rw_pointvec, WAND_SENSOR, VRDIR_FORE);
			vrUserTravelTranslate3d(VR_ALLUSERS,
				wand_rw_pointvec.v[VR_X] * delta_move,
				wand_rw_pointvec.v[VR_Y] * delta_move,
				wand_rw_pointvec.v[VR_Z] * delta_move);
		}
	}


#ifdef USE_SOCKETS /* { */
	/*******************************/
	/* Handle Commands from socket */

	/* if no connection (and a listen socket is open), see if someone is calling */
	if (cmd_socket < 0 && listen_sock >= 0) {
		/* cmd_socket will be -1 when no one is connected */
		cmd_socket = vrSocketAnswer(listen_sock);
		if (cmd_socket > 0) {
			printf("New client connected.\n");
			vrSocketSendMsg(cmd_socket, "Your wish is my command >");
		}
	}

	/* if a connection, see if client sent a command.  If so, act on it */
	if (cmd_socket > 0) {
		result = vrSocketReadMsgClean(cmd_socket, buf, 128);
		switch (result) {
		case -1:
			break;
		case  0:
			printf("Client stopped communicating.\n");
			close(cmd_socket);
			cmd_socket = -1;
			break;
		default:
			printf("socket cmd = '%s'\n", buf);
			/* parse command here -- very simple-minded parsing */
			if (strcmp(buf, "quit") == 0) {
				/* set the world for termination */
				wd->quit_bit = 1;
			} else
			if (strcmp(buf, "help") == 0 || strcmp(buf, "?") == 0) {
				/* give usage information */
				vrSocketSendMsg(cmd_socket, help_msg);
			} else
			if (strcmp(buf, "close") == 0) {
				/* close socket connection */
				printf("Client ending communication.\n");
				close(cmd_socket);
				cmd_socket = -1;
			} else
			if (strcmp(buf, "reset") == 0) {
				/* make the world as it was in the beginning */
				init_objects(wd);
				vrUserTravelReset(VR_ALLUSERS);
			} else
			if (strcmp(buf, "button?") == 0) {
				/* report the user-interface button value */
				sprintf(msg, "The button is %s.\n", (wd->object[3].ui_button_value ? "on" : "off"));
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strcmp(buf, "button on") == 0) {
				/* turn the user-interface button on */
				wd->object[3].ui_button_value = 1;
			} else
			if (strcmp(buf, "button off") == 0) {
				/* turn the user-interface button on */
				wd->object[3].ui_button_value = 0;
			} else
			if (strcmp(buf, "radio?") == 0) {
				/* report the user-interface radio-button value */
				sprintf(msg, "The radio button is %d.\n", wd->object[6].ui_button_value);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strncmp(buf, "radio ", 6) == 0) {
				/* set the user-interface radio-button value */
				wd->object[6].ui_button_value = atoi(&(buf[6]));
			} else
			if (strcmp(buf, "slider?") == 0) {
				/* report the user-interface slider value */
				sprintf(msg, "The slider is at %.2f.\n", wd->object[4].ui_valuator_value);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strncmp(buf, "slider ", 7) == 0) {
				/* set the user-interface slider value */
				wd->object[4].ui_valuator_value = atof(&(buf[7]));
			} else
			if (strcmp(buf, "lever?") == 0) {
				/* report the user-interface lever value */
				sprintf(msg, "The lever is at %.2f.\n", wd->object[5].ui_valuator_value);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strncmp(buf, "lever ", 6) == 0) {
				/* set the user-interface lever value */
				wd->object[5].ui_valuator_value = atof(&(buf[6]));
			} else
			if (strcmp(buf, "joyst?") == 0) {
				/* report the user-interface joystick value */
				sprintf(msg, "The joystick is at %.2f, %.2f.\n", wd->object[7].ui_valuator_value, wd->object[7].ui_valuator_value);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strncmp(buf, "joyst1 ", 7) == 0) {
				/* set the user-interface joystick value */
				wd->object[7].ui_valuator_value = atof(&(buf[7]));
			} else
			if (strncmp(buf, "joyst2 ", 7) == 0) {
				/* set the user-interface joystick value */
				wd->object[7].ui_valuator2_value = atof(&(buf[7]));
			} else
			if (strcmp(buf, "move cube up") == 0) {
				/* move the blue cube up 2.0 units */
				wd->object[0].y += 2.0;
			} else
			if (strcmp(buf, "move cube down") == 0) {
				/* move the blue cube down 2.0 units */
				wd->object[0].y -= 2.0;
			} else
			if (strcmp(buf, "where") == 0 || strcmp(buf, "where am i") == 0) {
				/* report the user's location */
				vrMatrix	*headmat;
				vrPoint		headloc;

				headmat = vrMatrixCreate();
				vrMatrixGetRWFromUserHead(headmat, 0);
				vrPointGetVWFromUserMatrix(&headloc, 0, headmat);
				sprintf(msg, "You are at (%f %f %f).\n", headloc.v[VR_X], headloc.v[VR_Y], headloc.v[VR_Z]);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strcmp(buf, "show ActiveWalls") == 0) {
				/* report the type of graphics (VR) display */
				sprintf(msg, "Application has %d active walls.\n", vrContext->config->num_windows);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			/* default: */ {
				vrSocketSendMsg(cmd_socket, "Unknown command.\n");
			}

			/* prompt for another command (if socket still open) */
			if (cmd_socket > 0) {
				vrSocketSendMsg(cmd_socket, "Your wish is my command >");
			}
		}
	}
#endif /* } USE_SOCKETS */

	/*********************************/
	/* Handle Commands from keyboard */
	if (vrGet2switchValue(0)) {
		wd->quit_bit = 1;
	}

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from wrs_shapes.c */
void	draw_floor();
void	draw_cube();
void	draw_pyramid();
void	draw_cube_outline();
void	draw_pyramid_outline();
void	draw_box_outlineonly(float x2, float x1, float y2, float y1, float z2, float z1);


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx()
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);
}


/*******************************************************************/
/* draw_world(): Now will render how ever many objects are in the  */
/*   world data structure.  The color, type (and maybe rotation)   */
/*   are all now stored for each object, in addition to location.  */
/*******************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */
	int		objnum;				/* object counter */
	ObjectType	*object;			/* pointer to current object */
	float		size;				/* size of current object */

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/************************************/
	/* draw a short pointer at the wand */
	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glScalef(0.15, 0.15, 0.15);
		glTranslatef(0.0, -1.0, 0.0);	/* move the tip to (0,0,0) */
		draw_pyramid_outline();
	glPopMatrix();

	/********************************************************/
	/* draw all the objects, taking care to put each one in */
	/*   the proper coordinate system (virtual world space  */
	/*   vs. real world space for the user-interface).      */

	/* draw the floor */
	glPushMatrix();
		vrRenderTransformUserTravel(rendinfo);	/* change to virtual-world coordinates */
		draw_floor();
	glPopMatrix();

	vrLockReadSet(wd->lock);	/* lock while reading info about objects */
	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);

		glColor3ubv(object->color);
		glPushMatrix();
			size = object->size;

			switch (object->type) {

			case vw_normal: /* render normal object (in virtual world coordinates) */
				vrRenderTransformUserTravel(rendinfo);	/* change to virtual-world coordinates */
				glTranslatef(object->x, object->y, object->z);
				glScalef(size, size, size);

				switch (object->shape) {
				case cube:
					if (objnum == wd->contacted_object)
						draw_cube_outline();
					else	draw_cube();
					break;
				case sphere:
					if (objnum == wd->contacted_object)
						draw_sphere_outline();
					else	draw_sphere();
					break;
				case pyramid:
					if (objnum == wd->contacted_object)
						draw_pyramid_outline();
					else	draw_pyramid();
					break;
				case ud_pyramid:
					glScalef(1.0, -1.0, 1.0);
					if (objnum == wd->contacted_object)
						draw_pyramid_outline();
					else	draw_pyramid();
					break;
				}
				break;

			case ui_button: /* render button widget (in real world coordinates) */
				/* widgets remain in real-world coordinates */
				glTranslatef(object->x, object->y, object->z);
#ifdef DRAW_UITEXT /* { */
				glPushMatrix();
					glTranslatef(0.0, -(size + 0.5), 0.0);

					glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
					if (object->ui_button_value)
						vrRenderText(rendinfo, "on");
					else	vrRenderText(rendinfo, "off");
				glPopMatrix();
#endif /* } DRAW_UITEXT */
				glScalef(size, size, size);
				glPushMatrix();

					/* for now we'll ignore shape and all buttons will look the same */

					/* draw the base box */
					if (objnum == wd->contacted_object)
						draw_cube_outline();
					else	draw_cube();

					/* draw the button part */
					if (object->ui_button_value) {
						/* button is pressed (so barely sticks out) */

						glTranslatef(0.4, 0.0, 0.0);
						/* move color halfway to white */
						glColor3ub(	
							object->color[0] + ((255 - object->color[0])/2),
							object->color[1] + ((255 - object->color[1])/2),
							object->color[2] + ((255 - object->color[2])/2));
					} else {
						/* button is not pressed (so sticks out more) */

						glTranslatef(0.8, 0.0, 0.0);
						/* move color halfway to black */
						glColor3ub(object->color[0]/2, object->color[1]/2, object->color[2]/2);
					}
					glScalef(0.7, 0.7, 0.7);
					draw_cube_outline();
				glPopMatrix();

				break;

			case ui_radio: { /* render radio-button widget (in real world coordinates)*/
				int	count;

				glTranslatef(object->x, object->y, object->z);
				glScalef(size, size, size);

				/* for now we'll ignore shape and all buttons will look the same */
				for (count = 0; count < object->ui_num_buttons; count++) {
					glPushMatrix();
						glTranslatef(0.0, ((object->ui_num_buttons - 1)/2.0 * (1.00 * (2.0 + RADIO_SEP))) - count * (1.00 * (2.0 + RADIO_SEP)), 0.0); /* yo */

						/* draw the base box */
						glColor3ubv(object->color);
						glScalef(1.0, 1.0, 1.0);
						if (objnum == wd->contacted_object)
							draw_cube_outline();
						else	draw_cube();

						/* draw the button part */
						if (object->ui_button_value == count) {
							/* this button is pressed (so barely sticks out) */

							glTranslatef(0.4, 0.0, 0.0);
							/* move color halfway to white */
							glColor3ub(	
								object->color[0] + ((255 - object->color[0])/2),
								object->color[1] + ((255 - object->color[1])/2),
								object->color[2] + ((255 - object->color[2])/2));
						} else {
							/* this button is not pressed (so sticks out more) */

							glTranslatef(0.8, 0.0, 0.0);
							/* move color halfway to black */
							glColor3ub(object->color[0]/2, object->color[1]/2, object->color[2]/2);
						}
#ifdef DRAW_UITEXT /* { */
						glPushMatrix();
						glTranslatef(0.0, 0.0, -1.5);
						glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
						vrRenderText(rendinfo, object->ui_radio_names[count]);
						glPopMatrix();
#endif /* } DRAW_UITEXT */
						glScalef(0.7, 0.7, 0.7);
						draw_cube_outline();
					glPopMatrix();
				}
			  }
				break;

			case ui_slider: /* render slider widget (in real world coordinates) */
				glTranslatef(object->x, object->y, object->z);
#ifdef DRAW_UITEXT /* { */
				glPushMatrix();
					glTranslatef(0.0, -SLIDER_LENGTH * size * 0.5 - 1.5, 0.0);

					sprintf(msg, "%5.2f", object->ui_valuator_value);
					glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
					vrRenderText(rendinfo, msg);
				glPopMatrix();
#endif /* } DRAW_UITEXT */
				glScalef(size, size, size);

				/* for now we'll ignore shape and all sliders will look the same */
				glPushMatrix();
					/* the sliding rod */
					glScalef(0.2, SLIDER_LENGTH, 0.2);
					if (objnum == wd->contacted_object)
						draw_cube_outline();
					else	draw_cube();
				glPopMatrix();

				glPushMatrix();
					/* the value marker */
					glColor3ub(object->color[0]/2, object->color[1]/2, object->color[2]/2);
					glTranslatef(0.0, object->ui_valuator_value * SLIDER_LENGTH, 0.0);
					glScalef(1.0, 0.2, 1.0);
					draw_cube_outline();
				glPopMatrix();
				glPushMatrix();
					glTranslatef(0.0, object->ui_valuator_value * SLIDER_LENGTH, 0.0);
					glScalef(0.201, 0.201, 0.201);
					draw_cube_outline();
				glPopMatrix();

				break;

			case ui_lever: /* render lever widget (in real world coordinates) */
				glTranslatef(object->x, object->y, object->z);
#ifdef DRAW_UITEXT /* { */
				glPushMatrix();
					glTranslatef(0.0, -1.0 * size - 0.5, 0.0);

					sprintf(msg, "%5.2f", object->ui_valuator_value);
					glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
					vrRenderText(rendinfo, msg);
				glPopMatrix();
#endif /* } DRAW_UITEXT */
				glScalef(size, size, size);

				/* for now we'll ignore shape and all sliders will look the same */
				glPushMatrix();
					/* the lever base */
					glScalef(1.0, 0.5, 0.2);
					if (objnum == wd->contacted_object)
						draw_cube_outline();
					else	draw_cube();
				glPopMatrix();

				glPushMatrix();
					/* the lever */
					glColor3ub(object->color[0]/2, object->color[1]/2, object->color[2]/2);
					glRotatef(90.0, 1.0, 0.0, 0.0);

					glRotatef(object->ui_valuator_value * -LEVER_MAX_ANGLE, 0.0, 0.0, 1.0);
					glTranslatef(0.0, LEVER_ARM_LENGTH, 0.0);
					glPushMatrix();
						glScalef(0.1, LEVER_ARM_LENGTH, 0.4);
						draw_pyramid_outline();
					glPopMatrix();

					/* a knob at the end of the lever */
					glColor3ubv(wd->object[objnum].color);
					glPushMatrix();
						glTranslatef(0.0, LEVER_ARM_LENGTH, 0.0);
						glScalef(LEVER_KNOB_SIZE, LEVER_KNOB_SIZE, LEVER_KNOB_SIZE);
						draw_sphere_outline();
					glPopMatrix();
				glPopMatrix();
				break;

			case ui_joyst: /* render joystick widget (in real world coordinates) */
				/* similar to lever, but the stick is up (+y) instead of +z */
				glTranslatef(object->x, object->y, object->z);
#ifdef DRAW_UITEXT /* { */
				glPushMatrix();
					glTranslatef(0.0, -1.0 * size - 0.5, 0.0);

					sprintf(msg, "%5.2f, %5.2f", object->ui_valuator_value, object->ui_valuator2_value);
					glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
					vrRenderText(rendinfo, msg);
				glPopMatrix();
#endif /* } DRAW_UITEXT */
				glScalef(size, size, size);

				/* for now we'll ignore shape and all sliders will look the same */
				glPushMatrix();
					/* the lever base */
					glScalef(1.0, 0.2, 1.0);
					if (objnum == wd->contacted_object)
						draw_cube_outline();
					else	draw_cube();
				glPopMatrix();

				glPushMatrix();
					/* the lever */
					glColor3ub(object->color[0]/2, object->color[1]/2, object->color[2]/2);
					glRotatef(000.0, 1.0, 0.0, 0.0);

					glRotatef(object->ui_valuator_value  * -JOYST_MAX_ANGLE, 0.0, 0.0, 1.0);
					glRotatef(object->ui_valuator2_value * -JOYST_MAX_ANGLE, 1.0, 0.0, 0.0);
					glTranslatef(0.0, JOYST_ARM_LENGTH, 0.0);
					glPushMatrix();
						glScalef(0.15, -JOYST_ARM_LENGTH, 0.15);
						draw_pyramid_outline();
					glPopMatrix();

					/* a knob at the end of the lever */
					glColor3ubv(wd->object[objnum].color);
					glPushMatrix();
						glTranslatef(0.0, JOYST_ARM_LENGTH, 0.0);
						glScalef(JOYST_KNOB_SIZE, JOYST_KNOB_SIZE, JOYST_KNOB_SIZE);
						draw_sphere_outline();
					glPopMatrix();
				glPopMatrix();
				break;
			}
		glPopMatrix();
	}
	/* NOTE: 2-D text is best rendered in a fixed location in the world */

	/*************************************************/
	/* draw text indicating which object is contacted */
	if (wd->contacted_object < 0) {
		sprintf(msg, "Object contacted: none");
	} else {
		sprintf(msg, "Object contacted: %d", wd->contacted_object);
	}

	vrLockReadRelease(wd->lock);

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 7.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

