/***********************************************************************/
/* FreeVR OpenGL tutorial example 6: ex6_text.c                        */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: January 11, 2007                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that        */
/*   reports some of the internal world data textually via a very      */
/*   simple 2D/raster font.  (We'll do 3D text in a later example.)    */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - FreeVR Library functions, variables, and macros:                */
/*	* type vrRenderInfo                                            */
/*      * vrRenderText(rendinfo, )                                     */
/*                                                                     */
/*                                                                     */
/* Considerations:                                                     */
/*   - This is our first exposure to the vrRenderInfo type.  But we    */
/*       will be using it in most of the rest of the examples.  For    */
/*       the most part, the application programmer does not need to    */
/*       worry about the details of this type & renderer argument.     */
/*       It can best be thought of as simply a conduit through which   */
/*       internal VR state information can be passed from the          */
/*       overarching render loop process to VR rendering operations    */
/*       that require this external state knowledge to work properly.  */
/*   - Textual output is often an important means of communicating     */
/*       with the user.  Certainly not every VR application requires   */
/*       the ability to put text about the state of the world in front */
/*       of the user.  However, it is always a useful feature to have  */
/*       when debugging a VR application (since the user/tester will   */
/*       generally not be sitting at the terminal to see printf-style  */
/*       messages).                                                    */
/*   - 2D raster (aka bitmap) text is generally less useful in a       */
/*       virtual reality experience, because the bitmap is a fixed     */
/*       size (number of pixels), and will be rendered the same no     */
/*       matter how close the user gets to it.  The best way to        */
/*       overcome this shortcoming is to always place the text at a    */
/*       location where this won't be noticed by the user.  In the     */
/*       case of a CAVE-like environment, this means placing the text  */
/*       to be located on the plane of one of the screens.  For a      */
/*       head-based system, the text should be on a head-up type       */
/*       widget that always remains a fixed distance from the user.    */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for sin(), cos(), fabs() & copysign() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		float		obj1_x, obj1_y, obj1_z;		/* location of object 1 */
		float		obj2_x, obj2_y, obj2_z;		/* location of object 2 */
		float		obj3_x, obj3_y, obj3_z;		/* location of object 3 */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations   */

void	init_gfx(void);				/* initialize the graphics         */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world */
void	init_world(WorldDataType *wd);		/* initialize the world parameters */
void	update_world(WorldDataType *wd);	/* perform the world simulation    */


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("FreeVR tutorial: example 6 -- simple text");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to show basic 2D text rendering");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Raise the green pyramid");
	vrInputSet2switchDescription(2, "Reset the green pyramid");
	vrInputSet2switchDescription(3, "Lower the green pyramid");
	vrInputSetValuatorDescription(0, "Move the green pyramid in the Z-plane");
	vrInputSetValuatorDescription(1, "Move the green pyramid in the X-plane");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* run until terminate key is pressed (Escape by default) */
	while(!vrGet2switchValue(0)) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/


/**************************************************************/
/* init_world(): create the initial conditions of the dynamic */
/*   simulation.                                              */
/**************************************************************/
void init_world(WorldDataType *wd)
{
	/* initial location of object 1 - a blue cube */
	wd->obj1_x = -2.0;
	wd->obj1_y =  5.0;
	wd->obj1_z = -5.0;

	/* initial location of object 2 - a red pyramid */
	wd->obj2_x =  2.0;
	wd->obj2_y =  5.0;
	wd->obj2_z = -5.0;

	/* initial location of object 3 - an upside-down green pyramid */
	wd->obj3_x =  0.0;
	wd->obj3_y =  5.0;
	wd->obj3_z = -5.0;
}


/*********************************************************************/
/* update_world(): Basic world with simple interactions.             */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a movement scale factor to tune control of pyramid   */
static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
	double		delta_time;			/* time since last update (in seconds) */
	double		joy_x, joy_y;			/* Joystick values */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/****************************************/
	/* Move the objects with constant speed */

	/* object 1 travels in a horizontal circle at a fixed speed */
	wd->obj1_x = 3*sin(sim_time * 2.0);
	wd->obj1_z = 3*cos(sim_time * 2.0) - 5.0;

	/* object 2 travels in a verticle circle at a fixed speed */
	wd->obj2_x = 3*cos(sim_time * 2.0);
	wd->obj2_y = 3*sin(sim_time * 2.0) + 5.0;

	/**************************/
	/* User interface section */

	/* the first button moves the object up, when initially pressed */
	if (vrGet2switchDelta(1) == 1) {
		wd->obj3_y += 1.0;
	}

	/* the second button resets the object's location, when initially pressed */
	if (vrGet2switchDelta(2) == 1) {
		wd->obj3_x =  0.0;
		wd->obj3_y =  5.0;
		wd->obj3_z = -5.0;
	}

	/* the third button moves the object down, when initially pressed */
	if (vrGet2switchDelta(3) == 1) {
		wd->obj3_y -= 1.0;
	}

	/* store input values so they are consistent throughout update routine */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* now use those values to move the object */
	if (fabs(joy_x) > JS_EPSILON)
		wd->obj3_x += joy_x * delta_time * MOVE_FACTOR;				/* NOTE: epsilon is NOT subtracted out */

	if (fabs(joy_y) > JS_EPSILON)
		wd->obj3_z -= (joy_y-copysign(JS_EPSILON,joy_y)) * delta_time * MOVE_FACTOR;/* NOTE: epsilon IS subtracted out */

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from shapes.c */
void	draw_cube();
void	draw_pyramid();


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx(void)
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */
}


/*************************************************************/
/* draw_world(): Render the three objects in the world.      */
/*   Note that two of the objects (the blue cube, and red    */
/*   pyramid) get their locations from the world database    */
/*************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/************************/
	/* draw all the objects */

	/* a blue cube */
	glColor3ub(100, 100, 255);
	glPushMatrix();
		glTranslatef(wd->obj1_x, wd->obj1_y, wd->obj1_z);
		draw_cube();
	glPopMatrix();

	/* a red pyramid */
	glColor3ub(255, 100, 100);
	glPushMatrix();
		glTranslatef(wd->obj2_x, wd->obj2_y, wd->obj2_z);
		draw_pyramid();
	glPopMatrix();

	/* the green upside-down pyramid */
	glColor3ub(100, 255, 100);
	glPushMatrix();
		glTranslatef(wd->obj3_x, wd->obj3_y, wd->obj3_z);
		glScalef(1.0, -1.0, 1.0);
		draw_pyramid();
	glPopMatrix();

	/*************************************************/
	/* draw text indicating the location of object 3 */
	sprintf(msg, "Green pyramid is at %.2f %.2f %.2f", wd->obj3_x, wd->obj3_y, wd->obj3_z);

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 7.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

