/***********************************************************************/
/* FreeVR OpenGL tutorial example 9: ex9_contact.c                     */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: December 19, 2006                                    */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that        */
/*   allows the user to select an object by touching it with the wand. */
/*   The selected object is highlighted by a black outline.            */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - FreeVR Library functions, variables, and macros:                */
/*      * vrPointGetRWFrom6sensor()                                    */
/*      * VR_X, VR_Y & VR_Z                                            */
/*                                                                     */
/* Considerations:                                                     */
/*   - Objects must be within reach                                    */
/*   - Pointer should reflect the region of contact                    */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for fabs() & copysign() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		float				x, y, z;	/* location of the object   */
		GLubyte				color[3];	/* color of the object      */
		enum { cube,pyramid,ud_pyramid }shape;		/* shape of the object      */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		number_of_objects;		/* number of objects in world */
		int		selected_object;		/* currently selected object  */
		ObjectType	object[3];			/* array of objects in world  */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations   */

void	init_gfx(void);				/* initialize the graphics         */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world */
void	init_world(WorldDataType *wd);		/* initialize the world parameters */
void	update_world(WorldDataType *wd);	/* perform the world simulation    */


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("example 9 -- contact");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to demonstrate contact between the pointer and objects");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Raise the green pyramid");
	vrInputSet2switchDescription(2, "Reset the green pyramid");
	vrInputSet2switchDescription(3, "Lower the green pyramid");
	vrInputSetValuatorDescription(0, "Move the green pyramid in the Z-plane");
	vrInputSetValuatorDescription(1, "Move the green pyramid in the X-plane");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a short cyan pointer to contact objects");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* run until terminate key is pressed (Escape by default) */
	while(!vrGet2switchValue(0)) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/


/**************************************************************/
/* init_world(): create the initial conditions of the dynamic */
/*   simulation.                                              */
/**************************************************************/
void init_world(WorldDataType *wd)
{
	int	objnum = 0;

	/* object 1 - a blue cube */
	wd->object[objnum].shape = cube;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x =  0.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 2 - a red pyramid */
	wd->object[objnum].shape = pyramid;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 3 - an upside-down green pyramid */
	wd->object[objnum].shape = ud_pyramid;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	wd->number_of_objects = objnum;
	wd->selected_object = -1;
}


/*********************************************************************/
/* update_world(): In this version, we handle the user interactions. */
/*   In this case, simply selecting objects in the world.            */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a movement scale factor to tune control of pyramid   */
static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
	vrPoint		wand_locpnt;			/* current location of the wand */
#define	wand_location wand_locpnt.v
	double		delta_time;			/* time since last update (in seconds) */
	double		joy_x, joy_y;			/* Joystick values */
	int		selected_obj;			/* the currently selected object  */
	int		touch_obj = -1;			/* the currently touched object   */
	int		objnum;				/* for looping throuh the objects */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/**************************/
	/* User interface section */

	/* the first button moves the object up, when initially pressed */
	if (vrGet2switchDelta(1) == 1) {
		wd->object[2].y += 1.0;
	}

	/* the second button resets the object's location, when initially pressed */
	if (vrGet2switchDelta(2) == 1) {
		wd->object[2].x =  4.0;
		wd->object[2].y =  5.0;
		wd->object[2].z = -4.0;
	}

	/* the third button moves the object down, when initially pressed */
	if (vrGet2switchDelta(3) == 1) {
		wd->object[2].y -= 1.0;
	}

	/* store input values so they are consistent throughout update routine */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* now use those values to move the object */
	if (fabs(joy_x) > JS_EPSILON)
		wd->object[2].x += joy_x * delta_time * MOVE_FACTOR;				/* NOTE: epsilon is NOT subtracted out */

	if (fabs(joy_y) > JS_EPSILON)
		wd->object[2].z -= (joy_y-copysign(JS_EPSILON,joy_y)) * delta_time * MOVE_FACTOR;/* NOTE: epsilon IS subtracted out */

	/**********************************/
	/** Handle Manipulation via wand **/
	/**********************************/

	/**********************************************************/
	/* check for contact between the wand and all the objects */

	vrPointGetRWFrom6sensor(&wand_locpnt, WAND_SENSOR);

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		if ( (wand_location[VR_X] < wd->object[objnum].x + 1.0)
		  && (wand_location[VR_X] > wd->object[objnum].x - 1.0)
		  && (wand_location[VR_Y] < wd->object[objnum].y + 1.0)
		  && (wand_location[VR_Y] > wd->object[objnum].y - 1.0)
		  && (wand_location[VR_Z] < wd->object[objnum].z + 1.0)
		  && (wand_location[VR_Z] > wd->object[objnum].z - 1.0) ) {
			touch_obj = objnum;
		}
	}

	selected_obj = wd->selected_object;

	if (selected_obj != touch_obj) {
		wd->selected_object = touch_obj;
	}

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from shapes.c */
void	draw_cube();
void	draw_pyramid();
void	draw_cube_outline();
void	draw_pyramid_outline();


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx(void)
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);
}


/*******************************************************************/
/* draw_world(): Now will render how ever many objects are in the  */
/*   world data structure.  The color, type (and maybe rotation)   */
/*   are all now stored for each object, in addition to location.  */
/*   The currently selected object is highlighted by adding a      */
/*   wireframe outline to the representation.                      */
/*******************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */
	int		objnum;				/* object counter */

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/************************************/
	/* draw a short pointer at the wand */
	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glScalef(0.15, 0.15, 0.15);
		glTranslatef(0.0, 1.5, 0.0);
		draw_pyramid_outline();
	glPopMatrix();

	/************************/
	/* draw all the objects */

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {

		glColor3ubv(wd->object[objnum].color);
		glPushMatrix();
			glTranslatef(wd->object[objnum].x, wd->object[objnum].y, wd->object[objnum].z);
			switch (wd->object[objnum].shape) {
			case cube:
				if (objnum == wd->selected_object)
					draw_cube_outline();
				else	draw_cube();
				break;
			case pyramid:
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			case ud_pyramid:
				glScalef(1.0, -1.0, 1.0);
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			}
		glPopMatrix();
	}

	/*************************************************/
	/* draw text indicating which object is selected */
	if (wd->selected_object < 0) {
		sprintf(msg, "Object selected: none");
	} else {
		sprintf(msg, "Object selected: %d", wd->selected_object);
	}

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 7.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

