/***********************************************************************/
/* FreeVR OpenGL tutorial example code: wrs_shapes.c                   */
/* Last Modified: August 19, 2010                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/*                                                                     */
/* This file contains routines to draw simple geometric shapes.        */
/*                                                                     */
/* The new features of this example include:                           */
/*   - new shapes are now included:                                    */
/*       - outline of a box (ie. parallelepiped)                       */
/*       - plane-segments                                              */
/*       - spheres (at different approximation levels                  */
/*   - texture coordinates are now included with all shapes.           */
/*   - declaration of "imagetype" for passing texture information      */
/*   - New routines to read/write PPM files:                           */
/*       - readPPM()                                                   */
/*       - writePPM3()                                                 */
/*   - Routines to algorithmically generate texture images:            */
/*       - make_texture_checker_pattern()                              */
/*   - A routine to set the OpenGL texture to a texture image:         */
/*       - set_texture_image()                                         */
/*                                                                     */
/* Considerations:                                                     */
/*   - by including the "set_texture_image()" routine within this      */
/*       source file, we can avoid the need to define the "imagetype"  */
/*       structure type outside this file.                             */
/*                                                                     */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>
#include <math.h>
#include <inttypes.h>		/* needed for "uint32_t" type */
#include <GL/gl.h>
#include <GL/glu.h>

#undef GLUT_DEBUG
#ifdef GLUT_DEBUG /* Include this if the any glut-text debugging section is enabled */
#include <GL/glut.h>
#endif

#include <freevr.h>		/* needed for vrShmemAlloc() */

void	draw_cube();
void	draw_cube_outline();
void	draw_pyramid();
void	draw_pyramid_outline();
void	draw_plane();
void	draw_plane_outline();
void	draw_plane_outline_only();
void	draw_sphere();
void	draw_sphere_outline();
void	enhance_sphere();
void	draw_enhanced_sphere();
void	draw_floor();


/*****************************************************************/
/*
   Cube (and pyramid) vertices.  Labeled 0-7 (cube) & 8 (top of pyramid).


             7_______ 4
             /|     /|        (+y)
            / | .8 / |         |
          0/__|___/3 |         |
           | 6|___|__|5        o---- (+x)
           | /    | /         /
           |/     |/         /
          1|______|2       (+z)

*/



float cube_vertices[9][3] = {			/* specifies the (x,y,z)   */
		{ -1.0,  1.0,  1.0 },	/* 0 */	/* location of each vertex */
		{ -1.0, -1.0,  1.0 },	/* 1 */
		{  1.0, -1.0,  1.0 },	/* 2 */
		{  1.0,  1.0,  1.0 },	/* 3 */
		{  1.0,  1.0, -1.0 },	/* 4 */
		{  1.0, -1.0, -1.0 },	/* 5 */
		{ -1.0, -1.0, -1.0 },	/* 6 */
		{ -1.0,  1.0, -1.0 },	/* 7 */
		{  0.0,  1.0,  0.0 },	/* 8 */
	};

int cube_polygons[6][4] = {			/* specifices the vertices */
		{ 0, 1, 2, 3 },	/* front  */	/* of each polygon         */
		{ 3, 2, 5, 4 },	/* right  */
		{ 5, 6, 7, 4 },	/* back   */
		{ 6, 1, 0, 7 },	/* left   */
		{ 5, 2, 1, 6 },	/* bottom */
		{ 0, 3, 4, 7 },	/* top    */
	};

float cube_normals[6][3] = {			/* specifies the normal */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon      */
		{  1.0,  0.0,  0.0 }, /* right  */
		{  0.0,  0.0, -1.0 }, /* back   */
		{ -1.0,  0.0,  0.0 }, /* left   */
		{  0.0, -1.0,  0.0 }, /* bottom */
		{  0.0,  1.0,  0.0 }, /* top    */
	};

int pyramid_polygons[5][4] = {			/* specifices the vertices */
		{ 1, 2, 8 },	/* front  */	/* of each polygon         */
		{ 2, 5, 8 },	/* right  */
		{ 5, 6, 8 },	/* back   */
		{ 6, 1, 8 },	/* left   */
		{ 6, 5, 2, 1 },	/* bottom */
	};

float pyramid_normals[5][3] = {			/* specifies the normal    */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon -- only */
		{  1.0,  0.0,  0.0 }, /* right  ** the bottom normal is    */
		{  0.0,  0.0, -1.0 }, /* back   ** perpendicular to the    */
		{ -1.0,  0.0,  0.0 }, /* left   ** surface.                */
		{  0.0, -1.0,  0.0 }, /* bottom */
	};

/* Plane vertices are the same as the first four cube vertices with z set to 0 */
float plane_vertices[4][3] = {			/* specifies the (x,y,z)   */
		{ -1.0,  1.0,  0.0 },	/* 0 */	/* location of each vertex */
		{ -1.0, -1.0,  0.0 },	/* 1 */
		{  1.0, -1.0,  0.0 },	/* 2 */
		{  1.0,  1.0,  0.0 },	/* 3 */
	};

double sphere_vertices[12][3] = {
		{  0.0                   ,  0.0                   ,  1.0                 }, /*  0 */
		{  0.89442719099991587856,  0.0                   ,  0.44721359549995793 }, /*  1 */
		{  0.27639320225002104342,  0.85065080835203993366,  0.44721359549995793 }, /*  2 */
		{ -0.72360679774997893378,  0.52573111211913365982,  0.44721359549995793 }, /*  3 */
		{ -0.72360679774997893378, -0.52573111211913365982,  0.44721359549995793 }, /*  4 */
		{  0.27639320225002104342, -0.85065080835203993366,  0.44721359549995793 }, /*  5 */
		{  0.72360679774997893378,  0.52573111211913365982, -0.44721359549995793 }, /*  6 */
		{ -0.27639320225002104342,  0.85065080835203993366, -0.44721359549995793 }, /*  7 */
		{ -0.89442719099991587856,  0.0                   , -0.44721359549995793 }, /*  8 */
		{ -0.27639320225002104342, -0.85065080835203993366, -0.44721359549995793 }, /*  9 */
		{  0.72360679774997893378, -0.52573111211913365982, -0.44721359549995793 }, /* 10 */
		{  0.0                   ,  0.0                   , -1.0                 }, /* 11 */
	};


int sphere_polygons[20][3] = {
		{  2,  0,  1 }, /* A */
		{  3,  0,  2 }, /* B */
		{  4,  0,  3 }, /* C */
		{  5,  0,  4 }, /* D */
		{  1,  0,  5 }, /* E */
		{  1,  6,  2 }, /* F */
		{  7,  2,  6 }, /* G */
		{  2,  7,  3 }, /* H */
		{  8,  3,  7 }, /* I */
		{  3,  8,  4 }, /* J */
		{  9,  4,  8 }, /* K */
		{  4,  9,  5 }, /* L */
		{ 10,  5,  9 }, /* M */
		{  6,  1, 10 }, /* N */
		{  5, 10,  1 }, /* O */
		{  6, 11,  7 }, /* P */
		{  7, 11,  8 }, /* Q */
		{  8, 11,  9 }, /* R */
		{  9, 11, 10 }, /* S */
		{ 10, 11,  6 }, /* T */
	};

double sphere_normals[20][3] = { { -10.0, -10.0, -10.0 } };	/* -10.0 indicates not set */

typedef	double	tritupled[3];
typedef	int	trituplei[3];
tritupled	*esphere_vertices;
trituplei	*esphere_polygons;
tritupled	*esphere_normals;
int		esphere_vextex_count;
int		esphere_poly_count;


/*****************************************************************/
/* NOTE: The texture coordinates are arranged such that each side of the cube  */
/*   gets the entire texture (an alternate method would be to have the texture */
/*   spread over the entire cube -- like is typically done with a sphere.      */
void draw_cube()
{
	int	poly;
	float	repeat_value = 1.0;

	for (poly = 0; poly < 6; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(cube_normals[poly]);
			glTexCoord2f(0.0, 0.0);
			glVertex3fv(cube_vertices[cube_polygons[poly][0]]);
			glTexCoord2f(0.0, repeat_value);
			glVertex3fv(cube_vertices[cube_polygons[poly][1]]);
			glTexCoord2f(repeat_value, repeat_value);
			glVertex3fv(cube_vertices[cube_polygons[poly][2]]);
			glTexCoord2f(repeat_value, 0.0);
			glVertex3fv(cube_vertices[cube_polygons[poly][3]]);
		glEnd();
	}
}


/*****************************************************************/
void draw_cube_outline()
{
	draw_cube();
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[5]);
	glEnd();
}


/******************************************************************/
/* Draw a box (parallelepiped) encapsulated by the 6 planes given */
/*   by the arguments.  Ie. x=x1 plane, x=x2 plane, y=y1 plane,   */
/*   etc.                                                         */
/* NOTE: which plane is minimum vs. maximum does not seem to      */
/*   matter, I tried all combinations, and the outline was drawn  */
/*   correctly -- note, only the outline has been tested thus far */
void draw_box_outlineonly(float x1, float x2, float y2, float y1, float z2, float z1)
{
	float	box_vertices[8][3];

	/* ordering is same as for cube above */
	box_vertices[0][0] = x1; box_vertices[0][1] = y2; box_vertices[0][2] = z1;
	box_vertices[1][0] = x1; box_vertices[1][1] = y1; box_vertices[1][2] = z1;
	box_vertices[2][0] = x2; box_vertices[2][1] = y1; box_vertices[2][2] = z1;
	box_vertices[3][0] = x2; box_vertices[3][1] = y2; box_vertices[3][2] = z1;
	box_vertices[4][0] = x2; box_vertices[4][1] = y2; box_vertices[4][2] = z2;
	box_vertices[5][0] = x2; box_vertices[5][1] = y1; box_vertices[5][2] = z2;
	box_vertices[6][0] = x1; box_vertices[6][1] = y1; box_vertices[6][2] = z2;
	box_vertices[7][0] = x1; box_vertices[7][1] = y2; box_vertices[7][2] = z2;

	glBegin(GL_LINE_STRIP);
		glVertex3fv(box_vertices[0]);
		glVertex3fv(box_vertices[1]);
		glVertex3fv(box_vertices[2]);
		glVertex3fv(box_vertices[3]);
		glVertex3fv(box_vertices[0]);
		glVertex3fv(box_vertices[7]);
		glVertex3fv(box_vertices[6]);
		glVertex3fv(box_vertices[1]);
		glVertex3fv(box_vertices[2]);
		glVertex3fv(box_vertices[5]);
		glVertex3fv(box_vertices[6]);
		glVertex3fv(box_vertices[7]);
		glVertex3fv(box_vertices[4]);
		glVertex3fv(box_vertices[3]);
		glVertex3fv(box_vertices[4]);
		glVertex3fv(box_vertices[5]);
	glEnd();
}


/*****************************************************************/
void draw_pyramid()
{
	int	poly;

	for (poly = 0; poly < 5; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(pyramid_normals[poly]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][0]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][1]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][2]]);
			if (poly == 4)
				glVertex3fv(cube_vertices[pyramid_polygons[poly][3]]);
		glEnd();
	}
}


/*****************************************************************/
void draw_pyramid_outline()
{
	draw_pyramid();
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
	glEnd();
}


/*****************************************************************/
/* A plane is just the front side of the cube ? */
void draw_plane()
{
	float	repeat_value = 1.0;

	glBegin(GL_POLYGON);
		glNormal3fv(cube_normals[0]);
		glTexCoord2f(0.0, 0.0);
		glVertex3fv(plane_vertices[1]);
		glTexCoord2f(repeat_value, 0.0);
		glVertex3fv(plane_vertices[2]);
		glTexCoord2f(repeat_value, repeat_value);
		glVertex3fv(plane_vertices[3]);
		glTexCoord2f(0.0, repeat_value);
		glVertex3fv(plane_vertices[0]);
	glEnd();

#ifdef GLUT_DEBUG /* render some GLUT text for debugging */
{
	char	msg[128] = "test msg";

	glPushAttrib(GL_COLOR_BUFFER_BIT);	/* should I include texture and lighting in this? */
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glColor3ub(100, 255, 100);
	glPushMatrix();
	glTranslated(plane_vertices[1][0], plane_vertices[1][1], plane_vertices[1][2]);
	glScaled(.0005, .0005, .0005);
	glRasterPos3f(0.0, 0.0, 0.0);
	sprintf(msg, "s:0.0 t:0.0");
	{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
	glPopMatrix();

	glPushMatrix();
	glTranslated(plane_vertices[2][0], plane_vertices[2][1], plane_vertices[2][2]);
	glScaled(.0005, .0005, .0005);
	glRasterPos3f(0.0, 0.0, 0.0);
	sprintf(msg, "s:1.0 t:0.0");
	{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
	glPopMatrix();

	glPushMatrix();
	glTranslated(plane_vertices[3][0], plane_vertices[3][1], plane_vertices[3][2]);
	glScaled(.0005, .0005, .0005);
	glRasterPos3f(0.0, 0.0, 0.0);
	sprintf(msg, "s:1.0 t:1.0");
	{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
	glPopMatrix();

	glPushMatrix();
	glTranslated(plane_vertices[0][0], plane_vertices[0][1], plane_vertices[0][2]);
	glScaled(.0005, .0005, .0005);
	glRasterPos3f(0.0, 0.0, 0.0);
	sprintf(msg, "s:0.0 t:1.0");
	{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
	glPopMatrix();

	/* restore the rendering settings */
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glPopAttrib();
}
#endif
}


/*****************************************************************/
void draw_plane_outline()
{
	draw_plane();
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(plane_vertices[0]);
		glVertex3fv(plane_vertices[1]);
		glVertex3fv(plane_vertices[2]);
		glVertex3fv(plane_vertices[3]);
		glVertex3fv(plane_vertices[0]);
	glEnd();
}


/*****************************************************************/
void draw_plane_outline_only()
{
	glBegin(GL_LINE_STRIP);
		glVertex3fv(plane_vertices[0]);
		glVertex3fv(plane_vertices[1]);
		glVertex3fv(plane_vertices[2]);
		glVertex3fv(plane_vertices[3]);
		glVertex3fv(plane_vertices[0]);
	glEnd();
}


/*****************************************************************/
void calc_sphere_normals()
{
	int	poly;

	for (poly = 0; poly < 20; poly++) {
		/* a simple normal is to average the three vertices */
		sphere_normals[poly][0] = sphere_vertices[sphere_polygons[poly][0]][0];
		sphere_normals[poly][1] = sphere_vertices[sphere_polygons[poly][0]][1];
		sphere_normals[poly][2] = sphere_vertices[sphere_polygons[poly][0]][2];

		sphere_normals[poly][0] += sphere_vertices[sphere_polygons[poly][1]][0];
		sphere_normals[poly][1] += sphere_vertices[sphere_polygons[poly][1]][1];
		sphere_normals[poly][2] += sphere_vertices[sphere_polygons[poly][1]][2];

		sphere_normals[poly][0] += sphere_vertices[sphere_polygons[poly][2]][0];
		sphere_normals[poly][1] += sphere_vertices[sphere_polygons[poly][2]][1];
		sphere_normals[poly][2] += sphere_vertices[sphere_polygons[poly][2]][2];

		sphere_normals[poly][0] /= 3.0;
		sphere_normals[poly][1] /= 3.0;
		sphere_normals[poly][2] /= 3.0;
	}
}


/*****************************************************************/
void calc_sphere_texcoord(double *coords, double *vertex)
/* NOTE: while this routine does give correct coordinate values, in fact it could */
/*   give better values if it had a little context for the overall polygon the vertex */
/*   is connected to.  For most of the problem, the solution was to add 1.0 to the 's' */
/*   coordinate if it was deemed that the polygon was really at the higher end of the  */
/*   texture.  However, another issue that has not been corrected, but could be corrected */
/*   with context information is for the vertices that is exactly at the north or south */
/*   pole of the sphere.  This method caluclates the 's' coordinate as '0.5'.  Which is */
/*   acceptible, but a better answer would be the average (or perhaps weighted average) */
/*   of the 's' coordinate from the other two vertices in the triangle.                 */
/* NOTE: [08/19/2010] Reversing s & t based on that was a fix implemented */
/*   in an experimental version in January of 2006.                       */
{
	double	s,t;

#if 1
	/* The method I was trying to calculate */
	t = -(vertex[2] + 1.0) / 2.0;	/* My first attempt -- this exagerates at the poles and squishes the equator */

	t = acos(vertex[2]) / M_PI;	/* taken from another example */
	s = -atan2(vertex[0], vertex[1]) / (M_PI * 2.0) + 0.5;
#endif

#if 0
	/* from P. Bourke webpage */
	t = acos(vertex[2]) / M_PI;
	s = acos(vertex[0] / sin(t * M_PI)) / (2.0 * M_PI);
	if (vertex[1] > 0.0)
		s += 0.5 * M_PI;
	s /= 2.0;
#endif

#if 0
	/* from http://www.mvps.org/directx/articles/spheremap.htm */
	t = asin(vertex[0]) / M_PI + 0.5;
	s = asin(vertex[1]) / M_PI + 0.5;
#endif

#if 0
	/* from http://msdn.microsoft.com/coding4fun/zman/zmanTextures3/default.aspx */
	t = acos(vertex[2]) / M_PI;
	if (vertex[2] == 1.0 || vertex[2] == -1.0) {
		s = 0.5;
	} else {
		s = acos(vertex[1] / sin(t * M_PI)) / (2.0 * M_PI);
		if (vertex[0] > 0.0)
			s = 1.0 - s;
	}
#endif



	coords[0] = s;
	coords[1] = t;
}


/*****************************************************************/
void draw_sphere()
{
	int	poly;
	double	*vertex1;
	double	*vertex2;
	double	*vertex3;
	double	texcoord1[2];
	double	texcoord2[2];
	double	texcoord3[2];

	if (sphere_normals[0][0] == -10.0)
		calc_sphere_normals();

	for (poly = 0; poly < 20; poly++) {
		vertex1 = sphere_vertices[sphere_polygons[poly][0]];
		vertex2 = sphere_vertices[sphere_polygons[poly][1]];
		vertex3 = sphere_vertices[sphere_polygons[poly][2]];
		calc_sphere_texcoord(texcoord1, vertex1);
		calc_sphere_texcoord(texcoord2, vertex2);
		calc_sphere_texcoord(texcoord3, vertex3);
		glBegin(GL_POLYGON);
			glNormal3dv(sphere_normals[poly]);
			glTexCoord2dv(texcoord1);
			glVertex3dv(vertex1);
			glTexCoord2dv(texcoord2);
			glVertex3dv(vertex2);
			glTexCoord2dv(texcoord3);
			glVertex3dv(vertex3);
		glEnd();

#ifdef GLUT_DEBUG /* render some GLUT text for debugging */
{
	char	msg[128] = "test msg";

		glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT);	/* should I include texture and lighting in this? */
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);

		glColor3ub(100, 255, 100);
		glPushMatrix();
		glTranslated(vertex1[0], vertex1[1], vertex1[2]);
		glScaled(.001, .001, .001);
		glRasterPos3f(0.0, 0.0, 0.0);
		sprintf(msg, "s:%.3f  t:%.3f", texcoord1[0], texcoord1[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		glPushMatrix();
		glTranslated(vertex2[0], vertex2[1], vertex2[2]);
		glScaled(.001, .001, .001);
		glRasterPos3f(0.0, 0.0, 0.0);
		sprintf(msg, "s:%.3f  t:%.3f", texcoord2[0], texcoord2[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		glPushMatrix();
		glTranslated(vertex3[0], vertex3[1], vertex3[2]);
		glScaled(.001, .001, .001);
		glRasterPos3f(0.0, 0.0, 0.0);
		sprintf(msg, "s:%.3f  t:%.3f", texcoord3[0], texcoord3[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		/* restore the rendering settings */
#  if 0
		glEnable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
#  endif
		glPopAttrib();
}
#endif
	}
}


/*****************************************************************/
/* NOTE: this doesn't do quite as efficient a job as cube or pyramid */
void draw_sphere_outline()
{
	int	poly;

	draw_sphere();
	glColor3ub(0, 0, 0);
	for (poly = 0; poly < 20; poly++) {
		glBegin(GL_LINE_LOOP);
			glVertex3dv(sphere_vertices[sphere_polygons[poly][0]]);
			glVertex3dv(sphere_vertices[sphere_polygons[poly][1]]);
			glVertex3dv(sphere_vertices[sphere_polygons[poly][2]]);
		glEnd();
	}
}


/*****************************************************************/
void draw_enhanced_sphere()
{
	int	poly;
	double	*vertex1;
	double	*vertex2;
	double	*vertex3;
	double	texcoord1[2];
	double	texcoord2[2];
	double	texcoord3[2];

	if (esphere_vertices == NULL) {
		esphere_vertices = (tritupled *)sphere_vertices;
		esphere_polygons = (trituplei *)sphere_polygons;
		esphere_normals = (tritupled *)sphere_normals;
		esphere_vextex_count = 12;
		esphere_poly_count = 20;	/* this is also for normals */

		enhance_sphere();
		enhance_sphere();
#if 1
		enhance_sphere();
#endif
	}

	if (sphere_normals[0][0] == -10.0)
		calc_sphere_normals();

	for (poly = 0; poly < esphere_poly_count; poly++) {
		vertex1 = esphere_vertices[esphere_polygons[poly][0]];
		vertex2 = esphere_vertices[esphere_polygons[poly][1]];
		vertex3 = esphere_vertices[esphere_polygons[poly][2]];
		calc_sphere_texcoord(texcoord1, vertex1);
		calc_sphere_texcoord(texcoord2, vertex2);
		calc_sphere_texcoord(texcoord3, vertex3);

#if 1 /* { */
		/* We need to make sure that the 's' texture coordinate is non-monotonic */
		/*   The cleaner way is probably to determine the direction from one vertex */
		/*   to the next, and making sure that the 's' coordinate goes in the same  */
		/*   direction.  But for now, we'll just do a test to see of the change in  */
		/*   texture is too great to make sense, and if so, adjust one or two of    */
		/*   the vertex's texture coordinates.                                      */
		if (texcoord1[0] > texcoord2[0] + 0.75) {
			texcoord2[0] += 1.0;	/* cause a wrap around */
		}
		if (texcoord2[0] > texcoord1[0] + 0.75) {
			texcoord1[0] += 1.0;	/* cause a wrap around */
		}
		if (texcoord1[0] > texcoord3[0] + 0.75) {
			texcoord3[0] += 1.0;	/* cause a wrap around */
		}
		if (texcoord3[0] > texcoord1[0] + 0.75) {
			texcoord1[0] += 1.0;	/* cause a wrap around */
		}
		if (texcoord2[0] > texcoord3[0] + 0.75) {
			texcoord3[0] += 1.0;	/* cause a wrap around */
		}
		if (texcoord3[0] > texcoord2[0] + 0.75) {
			texcoord2[0] += 1.0;	/* cause a wrap around */
		}

		/* Another correction/improvement we can make is to adjust the value of the 's' */
		/*   coordinate for the cases where 't' is 0.0 or 1.0 (ie. the poles of the     */
		/*   sphere.  In these cases, it would be better for the 's' coordinate of the  */
		/*   "pole" vertex to be the average of the other two 's' coordinates.          */
		if (texcoord1[1] == 0.00 || texcoord1[1] == 1.00) {
			texcoord1[0] = (texcoord2[0] + texcoord3[0]) / 2.0;
		}
		if (texcoord2[1] == 0.00 || texcoord2[1] == 1.00) {
			texcoord2[0] = (texcoord1[0] + texcoord3[0]) / 2.0;
		}
		if (texcoord3[1] == 0.00 || texcoord3[1] == 1.00) {
			texcoord3[0] = (texcoord2[0] + texcoord1[0]) / 2.0;
		}
#endif /* } */
#if 1
		glBegin(GL_POLYGON);
#else
		glBegin(GL_LINE_LOOP);
#endif
			glNormal3dv(esphere_normals[poly]);
			glTexCoord2dv(texcoord1);
			glVertex3dv(vertex1);
			glTexCoord2dv(texcoord2);
			glVertex3dv(vertex2);
			glTexCoord2dv(texcoord3);
			glVertex3dv(vertex3);
		glEnd();

#ifdef GLUT_DEBUG /* render some GLUT text for debugging */
{
	char	msg[128] = "test msg";

		glPushAttrib(GL_COLOR_BUFFER_BIT);	/* should I include texture and lighting in this? */
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glColor3ub(100, 255, 100);
		glPushMatrix();
		glTranslated(vertex1[0], vertex1[1], vertex1[2]);
		glScaled(.0005, .0005, .0005);
		glRasterPos3f(0.0, 0.0, 0.0);
		//sprintf(msg, "s:%.3f  t:%.3f", texcoord1[0], texcoord1[1]);
		sprintf(msg, "s:%.3f", texcoord1[0]);
		//sprintf(msg, "t:%.3f", texcoord1[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		glPushMatrix();
		glTranslated(vertex2[0], vertex2[1], vertex2[2]);
		glScaled(.0005, .0005, .0005);
		glRasterPos3f(0.0, 0.0, 0.0);
		//sprintf(msg, "s:%.3f  t:%.3f", texcoord2[0], texcoord2[1]);
		sprintf(msg, "s:%.3f", texcoord2[0]);
		//sprintf(msg, "t:%.3f", texcoord2[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		glPushMatrix();
		glTranslated(vertex3[0], vertex3[1], vertex3[2]);
		glScaled(.0005, .0005, .0005);
		glRasterPos3f(0.0, 0.0, 0.0);
		//sprintf(msg, "s:%.3f  t:%.3f", texcoord3[0], texcoord3[1]);
		sprintf(msg, "s:%.3f", texcoord3[0]);
		//sprintf(msg, "t:%.3f", texcoord3[1]);
		{ char	*chr; for (chr = msg; *chr; chr++) glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr); }
		glPopMatrix();

		/* restore the rendering settings */
		glEnable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
		glPopAttrib();
}
#endif
	}
}


/*****************************************************************/
/* enhance_sphere(): subdivides the polygons on a sphere to make */
/*   a better spherical approximation out of 3-polygons (ie.     */
/*   triangles).                                                 */
/*                                                               */
/* Uses subdivision code provided by Stuart Levy                 */
/*   The initial triangle (v0, v1, v2) is divided into four new  */
/*   triangles, with three new vertices.                         */
/*          v1                                                   */
/*         /  \                                                  */
/*        /  0 \                                                 */
/*       v3----v5                                                */
/*      / \  2 / \                                               */
/*     / 1 \  / 3 \                                              */
/*    v2----v4----v0                                             */
/*****************************************************************/
#if 1 /* see slevy_sphere for the awk script that does this -- pretty simple */
/*****************************************************************/
void calc_new_vertex(tritupled *new_vert, tritupled vert1, tritupled vert2)
{
	double		radius = 1.0;	/* TODO: make this an argument */
	double		tx, ty, tz;
	double		adjust;

	tx = vert1[0] + vert2[0];
	ty = vert1[1] + vert2[1];
	tz = vert1[2] + vert2[2];
	adjust = radius / sqrt(tx*tx + ty*ty + tz*tz);

	(*new_vert)[0] = tx * adjust;
	(*new_vert)[1] = ty * adjust;
	(*new_vert)[2] = tz * adjust;

}


/*****************************************************************/
void enhance_sphere()
{
	double		radius = 1.0;	/* TODO: make this an argument */
	int		count;
	int		new_vertex_count = 0;
	int		new_poly_count = 0;
	tritupled	*new_esphere_vertices;
	trituplei	*new_esphere_polygons;
	tritupled	*new_esphere_normals;
	tritupled	vertex1;
	tritupled	vertex2;
	tritupled	vertex3;
	tritupled	new_vertex;
	int		vertex1_num;
	int		vertex2_num;
	int		vertex3_num;
	int		vertex4_num;
	int		vertex5_num;
	int		vertex6_num;

printf("enhance sphere!\n\n\n");
	if (esphere_vertices == NULL) {
printf("enhance_sphere(): initializing enhanced sphere\n");
		esphere_vertices = (tritupled *)sphere_vertices;
		esphere_polygons = (trituplei *)sphere_polygons;
		esphere_normals = (tritupled *)sphere_normals;
		esphere_vextex_count = 12;
		esphere_poly_count = 20;	/* this is also for normals */
	}

	new_esphere_vertices = (tritupled *)vrShmemAlloc(sizeof(tritupled) * (esphere_vextex_count + esphere_poly_count * 4));	/* TODO: figure out exactly how many there will be */
	new_esphere_polygons = (trituplei *)vrShmemAlloc(sizeof(trituplei) * esphere_poly_count * 4);
	new_esphere_normals  = (tritupled *)vrShmemAlloc(sizeof(tritupled) * esphere_poly_count * 4);

	/* first copy the existing vertices into the new list */
	for (count = 0; count < esphere_vextex_count; count++) {
		new_esphere_vertices[count][0] = esphere_vertices[count][0];
		new_esphere_vertices[count][1] = esphere_vertices[count][1];
		new_esphere_vertices[count][2] = esphere_vertices[count][2];
		new_vertex_count++;
	}

	/* loop over each face, adding the three new vertices, and four */
	/*   polygons (while removing the previous single polygon) to   */
	/*   the appropriate lists.                                     */
	/* NOTE: there is a lot of redundancy in the creation of new vertices that */
	/*   aren't shared between new polygons.                                   */

	for (count = 0; count < esphere_poly_count; count++) {
		vertex1_num = esphere_polygons[count][0];
		vertex2_num = esphere_polygons[count][1];
		vertex3_num = esphere_polygons[count][2];

		vertex1[0] = esphere_vertices[vertex1_num][0];
		vertex1[1] = esphere_vertices[vertex1_num][1];
		vertex1[2] = esphere_vertices[vertex1_num][2];

		vertex2[0] = esphere_vertices[vertex2_num][0];
		vertex2[1] = esphere_vertices[vertex2_num][1];
		vertex2[2] = esphere_vertices[vertex2_num][2];

		vertex3[0] = esphere_vertices[vertex3_num][0];
		vertex3[1] = esphere_vertices[vertex3_num][1];
		vertex3[2] = esphere_vertices[vertex3_num][2];

		/* Add the three new vertices */
		calc_new_vertex(&new_vertex, vertex2, vertex3); /* TODO: replace directly w/ new_esphere_vertices */
		new_esphere_vertices[new_vertex_count][0] = new_vertex[0];
		new_esphere_vertices[new_vertex_count][1] = new_vertex[1];
		new_esphere_vertices[new_vertex_count][2] = new_vertex[2];
		vertex4_num = new_vertex_count++;

		calc_new_vertex(&new_esphere_vertices[new_vertex_count], vertex3, vertex1);
		vertex5_num = new_vertex_count++;

		calc_new_vertex(&new_esphere_vertices[new_vertex_count], vertex1, vertex2);
		vertex6_num = new_vertex_count++;


printf("new poly %d -- v1 = %d, v2 = %d, v3 = %d, v4 = %d, v5 = 5d, v6 = %d\n", new_poly_count, vertex1_num, vertex2_num, vertex3_num, vertex4_num, vertex5_num, vertex6_num);
printf("reiterate: new_esphere_polygons[0][*] = %d,%d,%d\n", new_esphere_polygons[0][0], new_esphere_polygons[0][1], new_esphere_polygons[0][2]);
		/* Specify the four replacement polygons */
		new_esphere_polygons[new_poly_count][0] = vertex6_num;
		new_esphere_polygons[new_poly_count][1] = vertex2_num;
		new_esphere_polygons[new_poly_count][2] = vertex4_num;
printf("new poly %d -- new_esphere_polygons[npc][*] (%p) = %d,%d,%d, v6 = %d, v2 = %d, v4 = %d\n", new_poly_count, &(new_esphere_polygons[new_poly_count][0]), new_esphere_polygons[new_poly_count][0], new_esphere_polygons[new_poly_count][1], new_esphere_polygons[new_poly_count][2], vertex6_num, vertex2_num, vertex4_num);
		new_poly_count++;

		new_esphere_polygons[new_poly_count][0] = vertex5_num;
		new_esphere_polygons[new_poly_count][1] = vertex4_num;
		new_esphere_polygons[new_poly_count][2] = vertex3_num;
printf("new poly %d -- new_esphere_polygons[npc][*] (%p) = %d,%d,%d, v6 = %d, v2 = %d, v4 = %d\n", new_poly_count, &(new_esphere_polygons[new_poly_count][0]), new_esphere_polygons[new_poly_count][0], new_esphere_polygons[new_poly_count][1], new_esphere_polygons[new_poly_count][2], vertex5_num, vertex4_num, vertex3_num);
		new_poly_count++;
printf("reiterate: new_esphere_polygons[0][*] = %d,%d,%d\n", new_esphere_polygons[0][0], new_esphere_polygons[0][1], new_esphere_polygons[0][2]);

		new_esphere_polygons[new_poly_count][0] = vertex4_num;
		new_esphere_polygons[new_poly_count][1] = vertex5_num;
		new_esphere_polygons[new_poly_count][2] = vertex6_num;
printf("new poly %d -- new_esphere_polygons[npc][*] (%p) = %d,%d,%d, v6 = %d, v2 = %d, v4 = %d\n", new_poly_count, &(new_esphere_polygons[new_poly_count][0]), new_esphere_polygons[new_poly_count][0], new_esphere_polygons[new_poly_count][1], new_esphere_polygons[new_poly_count][2], vertex4_num, vertex5_num, vertex6_num);
		new_poly_count++;
printf("reiterate: new_esphere_polygons[0][*] = %d,%d,%d\n", new_esphere_polygons[0][0], new_esphere_polygons[0][1], new_esphere_polygons[0][2]);

		new_esphere_polygons[new_poly_count][0] = vertex1_num;
		new_esphere_polygons[new_poly_count][1] = vertex6_num;
		new_esphere_polygons[new_poly_count][2] = vertex5_num;
printf("new poly %d -- new_esphere_polygons[npc][*] (%p) = %d,%d,%d, v6 = %d, v2 = %d, v4 = %d\n", new_poly_count, &(new_esphere_polygons[new_poly_count][0]), new_esphere_polygons[new_poly_count][0], new_esphere_polygons[new_poly_count][1], new_esphere_polygons[new_poly_count][2], vertex1_num, vertex6_num, vertex5_num);
		new_poly_count++;
printf("reiterate: new_esphere_polygons[0][*] = %d,%d,%d\n", new_esphere_polygons[0][0], new_esphere_polygons[0][1], new_esphere_polygons[0][2]);
printf("nvc = %d, npc = %d\n", new_vertex_count, new_poly_count);
	}
printf("nvc = %d, npc = %d\n", new_vertex_count, new_poly_count);

	/* Calculate the new normals */
	for (count = 0; count < new_poly_count; count++) {
		/* a simple normal is to average the three vertices */
printf("cal normal %d -- new_esphere_polygons[count][*] (%p) = %d,%d,%d\n", count,
	&(new_esphere_polygons[count][0]),
	new_esphere_polygons[count][0],
	new_esphere_polygons[count][1],
	new_esphere_polygons[count][2]);
#if 0
		new_esphere_normals[count][0] = new_esphere_vertices[new_esphere_polygons[count][0]][0];
		new_esphere_normals[count][1] = new_esphere_vertices[new_esphere_polygons[count][0]][1];
		new_esphere_normals[count][2] = new_esphere_vertices[new_esphere_polygons[count][0]][2];

		new_esphere_normals[count][0] += new_esphere_vertices[new_esphere_polygons[count][1]][0];
		new_esphere_normals[count][1] += new_esphere_vertices[new_esphere_polygons[count][1]][1];
		new_esphere_normals[count][2] += new_esphere_vertices[new_esphere_polygons[count][1]][2];

		new_esphere_normals[count][0] += new_esphere_vertices[new_esphere_polygons[count][2]][0];
		new_esphere_normals[count][1] += new_esphere_vertices[new_esphere_polygons[count][2]][1];
		new_esphere_normals[count][2] += new_esphere_vertices[new_esphere_polygons[count][2]][2];

		new_esphere_normals[count][0] /= 3.0;
		new_esphere_normals[count][1] /= 3.0;
		new_esphere_normals[count][2] /= 3.0;
#endif
	}
printf("done calculating normals\n");


	/* Make a couple of verification checks */
	if (new_vertex_count != esphere_vextex_count * 2) {
		printf("something went wrong: new_vertex_count = %d, esphere_vextex_count * 2 = %d\n",
			new_vertex_count, esphere_vextex_count * 2);
	}
	if (new_poly_count != esphere_poly_count * 4) {
		printf("something went wrong: new_poly_count = %d, esphere_poly_count * 4 = %d\n",
			new_poly_count, esphere_poly_count * 4);
	}

	esphere_vertices = new_esphere_vertices;
	esphere_polygons = new_esphere_polygons;
	esphere_normals = new_esphere_normals;
	esphere_vextex_count = new_vertex_count;
	esphere_poly_count = new_poly_count;
	printf("enhanced sphere has %d vertices and %d polygons\n", esphere_vextex_count, esphere_poly_count);
}
#endif


/*****************************************************************/
/* draw_floor(): draws a floor especially designed for the CAVE. */
/*****************************************************************/
void draw_floor()
{
	glColor3ub(25, 25, 200);
	glBegin(GL_POLYGON);
		glNormal3f( 0.0, 1.0,  0.0);
		glVertex3f(-5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0,  5.0);
		glVertex3f(-5.0, 0.0,  5.0);
	glEnd();
}


/******************************************************************/
/******************************************************************/
/*  This section of this file could be moved to a textures.c file */
/******************************************************************/

typedef	struct {
		unsigned long	xsize;
		unsigned long	ysize;
		unsigned long	zsize;
		uint32_t	*pixels;
	} imageType;

imageType	*make_texture_checker_pattern(int xsize, int ysize, int trans);
imageType	*make_texture_random_pattern(int xsize, int ysize, int trans);
imageType	*readPPM(char *filename);
void		writePPM3(char *filename, imageType *image);
void		set_texture_image(imageType *image);


/**********************************************************************/
/* Make a checker pattern for use as a texture map.  Size is xsize by */
/*   ysize, and the color scheme is either GL-color vs. black, or     */
/*   GL-color vs. transparent hole.  Transparency is set by the trans */
/*   argument.                                                        */
/* NOTE: memory is allocated here from the shared memory pool, with   */
/*   a pointer to the newly allocated memory returned to the calling  */
/*   function.                                                        */
/* TODO: this needs to take an allocation function as an argument */
imageType *make_texture_checker_pattern(int xsize, int ysize, int trans)
{
	int		xcount;
	int		ycount;
	uint32_t	value1 = 0xFFFFFFFF;
	uint32_t	value2 = (trans ? 0x60606000 : 0x606060FF);
	imageType	*image;

	image = (imageType *)vrShmemAlloc(sizeof(imageType));
	image->pixels = (uint32_t *)vrShmemAlloc(sizeof(uint32_t) * xsize * ysize);
	image->xsize = xsize;
	image->ysize = ysize;
	image->zsize = 4;	/* 3 colors + opacity */

	for (xcount = 0; xcount < xsize; xcount++) {
		for (ycount = 0; ycount < ysize; ycount++) {
			image->pixels[ycount * xsize + xcount] = ((xcount%2 == ycount%2) ? value1 : value2);
		}
	}

	return image;
}



/**********************************************************************/
/* Make a random pattern for use as a texture map.  Size is xsize by  */
/*   ysize.  Transparency (the "trans" argument) is currently unused  */
/*   but kept for symetry with the checkerboard version.              */
imageType *make_texture_random_pattern(int xsize, int ysize, int trans)
{
	int		xcount;
	int		ycount;
	unsigned char	red;
	unsigned char	green;
	unsigned char	blue;
	unsigned char	opacity;
	imageType	*image;

	image = (imageType *)vrShmemAlloc(sizeof(imageType));
	image->pixels = (uint32_t *)vrShmemAlloc(sizeof(uint32_t) * xsize * ysize);
	image->xsize = xsize;
	image->ysize = ysize;
	image->zsize = 4;	/* 3 colors + opacity */

	for (xcount = 0; xcount < xsize; xcount++) {
		for (ycount = 0; ycount < ysize; ycount++) {
			red     = floor(drand48() * 255);
			green   = floor(drand48() * 255);
			blue    = floor(drand48() * 255);
			opacity = 255;

			image->pixels[ycount * xsize + xcount] = (red << 24) | (blue << 16) | (green << 8) | opacity;
		}
	}

	return image;
}


/*****************************************************************/
void writePPM3(char *filename, imageType *image)
{
	FILE		*fp;
	uint32_t	pixel;
	int		xcount;
	int		ycount;

	if (image == NULL) {
		fprintf(stderr, "writePPM3(): NULL image data\n");
		return;
	}

	fp = fopen(filename, "w+");

	if (fp == NULL) {
		fprintf(stderr, "Could not open '%s' for writing\n", filename);
		return;
	}

	fprintf(fp, "P3\n");
	fprintf(fp, "# Created by Bill Sherman's sample shapes.c file\n");
	fprintf(fp, "%d %d\n", image->xsize, image->ysize);
	fprintf(fp, "255\n");
	for (xcount = 0; xcount < image->xsize; xcount++) {
		for (ycount = 0; ycount < image->ysize; ycount++) {
			pixel = image->pixels[ycount * image->xsize + xcount];
			fprintf(fp, "%d %d %d\n", (pixel>>24)&0xFF, (pixel>>16)&0xFF, (pixel>>8)&0xFF);
		}
	}
	fclose(fp);
}


/*****************************************************************/
/* TODO: ideally, this would take an allocation function as an argument */
imageType *readPPM(char *filename)
{
	FILE		*fp;
	imageType	*image;
	uint32_t	pixel;
	int		xcount;
	int		ycount;
	int		ppm_type;
	int		width;
	int		height;
	int		maxvalue;
	unsigned short	red, green, blue;
	int		fscan_val;

	fp = fopen(filename, "r");

	if (fp == NULL) {
		fprintf(stderr, "readPPM(): Could not open '%s' for reading\n", filename);
		return NULL;
	}

	fscan_val = fscanf(fp, "P%d%*[ \t\n]", &ppm_type);	/* search to the P<n> + whitespace */
	if (fscan_val != 1) {
		fprintf(stderr, "readPPM(): '%s' does not seem to be a PPM image file\n", filename);
		return NULL;
	}

	printf("reading PPM type %d image\n", ppm_type);
	if (ppm_type != 6) {
		fprintf(stderr, "readPPM(): currently only able to ready PPM-6 files (%s)\n", filename);
		return NULL;
	}
#if 0 /* some files may have comments, so skip them */
{
	fscan_val = fscanf(fp, "#%*[\n]");			/* pass a line with comments */
	printf("fscan_val = %d\n", fscan_val);
	fscan_val = fscanf(fp, "#%*[\n]");			/* pass a line with comments */
	printf("fscan_val = %d\n", fscan_val);
}
#endif
	fscanf(fp, "%d", &width);
	fscanf(fp, "%d", &height);
	fscanf(fp, "%d", &maxvalue);
	fscanf(fp, "\n");
#if 1
printf("filename '%s': width = %d, height = %d, maxvalue = %d, imageTypesize is %d, size of array is %d\n", filename, width, height, maxvalue, sizeof(imageType), sizeof(uint32_t) * width * height);
#endif

	/* TODO: consider doing a power-of-2 test here */

	image = (imageType *)vrShmemAlloc(sizeof(imageType));
	image->xsize = width;
	image->ysize = height;
	image->pixels = (uint32_t *)vrShmemAlloc(sizeof(uint32_t) * image->xsize * image->ysize);

	for (xcount = 0; xcount < image->xsize; xcount++) {
		for (ycount = 0; ycount < image->ysize; ycount++) {
#if 0 /* this is for type-3 PPMs */
			fscanf(fp, "%d", &red);
			fscanf(fp, "%d", &green);
			fscanf(fp, "%d", &blue);
#else /* this is for type-6 PPMs */
			red   = (unsigned short)fgetc(fp);
			green = (unsigned short)fgetc(fp);
			blue  = (unsigned short)fgetc(fp);
#endif
			pixel = 0x000000FF;		/* alpha = full opacity */
			pixel |= (red  <<24) & 0xFF000000;
			pixel |= (green<<16) & 0x00FF0000;
			pixel |= (blue << 8) & 0x0000FF00;
			image->pixels[(image->xsize-xcount) * image->ysize + ycount] = pixel;
		}
	}
	fclose(fp);
printf("done reading image '%s'\n", filename);

	return (image);
}


/*****************************************************************/
void set_texture_image(imageType *image)
{
	if (image == NULL) {
		fprintf(stderr, "set_texture(): NULL image data\n");
		return;
	}

	/* TODO: consider doing a power-of-2 test here */

	glTexImage2D(GL_TEXTURE_2D,	/* GL_TEXTURE_2D or GL_PROXY_TEXTURE_2D */
			0,		/* level-of-detail (mip-map), 0 is base */
			4,		/* number of color components or format */
			image->xsize,	/* x-size of image */
			image->ysize,	/* y-size of image */
			0,		/* size of border (0 or 1) */
			GL_RGBA,	/* format of the pixel data */
			GL_UNSIGNED_INT_8_8_8_8,/* type of the pixel data */
			image->pixels);	/* pointer to the image data */
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	return;
}


